"""
Accessible models can be read and copied but not modified or deleted.

Owned models can be modified and deleted.
"""
from typing import (
    Any,
    Optional,
    Type,
    TYPE_CHECKING,
)

from galaxy import (
    exceptions,
    model,
)

if TYPE_CHECKING:
    from sqlalchemy.orm import Query


class AccessibleManagerMixin:
    """
    A security interface to check if a User can read/view an item's.

    This can also be thought of as 'read but not modify' privileges.
    """

    # declare what we are using from base ModelManager
    model_class: Type[Any]

    def by_id(self, id: int):
        ...

    # don't want to override by_id since consumers will also want to fetch w/o any security checks
    def is_accessible(self, item: "Query", user: model.User, **kwargs: Any) -> bool:
        """
        Return True if the item accessible to user.
        """
        # override in subclasses
        raise exceptions.NotImplemented("Abstract interface Method")

    def get_accessible(self, id: int, user: model.User, **kwargs: Any) -> "Query":
        """
        Return the item with the given id if it's accessible to user,
        otherwise raise an error.

        :raises exceptions.ItemAccessibilityException:
        """
        item = self.by_id(id)
        return self.error_unless_accessible(item, user, **kwargs)

    def error_unless_accessible(self, item: "Query", user, **kwargs):
        """
        Raise an error if the item is NOT accessible to user, otherwise return the item.

        :raises exceptions.ItemAccessibilityException:
        """
        if self.is_accessible(item, user, **kwargs):
            return item
        raise exceptions.ItemAccessibilityException(f"{self.model_class.__name__} is not accessible by user")

    # TODO:?? are these even useful?
    def list_accessible(self, user, **kwargs):
        """
        Return a list of items accessible to the user, raising an error if ANY
        are inaccessible.

        :raises exceptions.ItemAccessibilityException:
        """
        raise exceptions.NotImplemented("Abstract interface Method")
        # NOTE: this will be a large, inefficient list if filters are not passed in kwargs
        # items = ModelManager.list( self, trans, **kwargs )
        # return [ self.error_unless_accessible( trans, item, user ) for item in items ]

    def filter_accessible(self, user, **kwargs):
        """
        Return a list of items accessible to the user.
        """
        raise exceptions.NotImplemented("Abstract interface Method")
        # NOTE: this will be a large, inefficient list if filters are not  passed in kwargs
        # items = ModelManager.list( self, trans, **kwargs )
        # return filter( lambda item: self.is_accessible( trans, item, user ), items )


class OwnableManagerMixin:
    """
    A security interface to check if a User is an item's owner.

    Some resources are associated with the User that created or imported them
    and these Users can be considered the models' owner.

    This can also be thought of as write/edit privileges.
    """

    # declare what we are using from base ModelManager
    model_class: Type[Any]

    def by_id(self, id: int):
        ...

    def is_owner(self, item: model.Base, user: Optional[model.User], **kwargs: Any) -> bool:
        """
        Return True if user owns the item.
        """
        # override in subclasses
        raise exceptions.NotImplemented("Abstract interface Method")

    def get_owned(self, id: int, user: Optional[model.User], **kwargs: Any) -> Any:
        """
        Return the item with the given id if owned by the user,
        otherwise raise an error.

        :raises exceptions.ItemOwnershipException:
        """
        item = self.by_id(id)
        return self.error_unless_owner(item, user, **kwargs)

    def error_unless_owner(self, item, user: Optional[model.User], **kwargs: Any):
        """
        Raise an error if the item is NOT owned by user, otherwise return the item.

        :raises exceptions.ItemAccessibilityException:
        """
        if self.is_owner(item, user, **kwargs):
            return item
        raise exceptions.ItemOwnershipException(f"{self.model_class.__name__} is not owned by user")

    def list_owned(self, user, **kwargs):
        """
        Return a list of items owned by the user, raising an error if ANY
        are not.

        :raises exceptions.ItemAccessibilityException:
        """
        raise exceptions.NotImplemented("Abstract interface Method")
        # just alias to by_user (easier/same thing)
        # return self.by_user( trans, user, **kwargs )

    def filter_owned(self, user, **kwargs):
        """
        Return a list of items owned by the user.
        """
        # just alias to list_owned
        return self.list_owned(user, **kwargs)



import logging
from os.path import dirname

from galaxy.queue_worker import job_rule_modules
from galaxy.structured_app import StructuredApp
from galaxy.tool_util.toolbox.watcher import (
    get_tool_conf_watcher,
    get_tool_watcher,
)
from galaxy.util.watcher import get_watcher

log = logging.getLogger(__name__)


class ConfigWatchers:
    """Contains ToolConfWatcher, ToolWatcher and ToolDataWatcher objects."""

    def __init__(self, app: StructuredApp):
        self.app = app
        self.active = False
        # ToolConfWatcher objects will watch the tool_cache if the tool_cache is passed into get_tool_conf_watcher.
        # Watching the tool_cache means removing outdated items from the tool_cache.
        # Only the reload_toolbox callback will re-populate the cache, so we pass the tool_cache only to the ToolConfWatcher that
        # watches regular tools.
        # If there are multiple ToolConfWatcher objects for the same handler or web process a race condition occurs between the two cache_cleanup functions.
        # If the reload_data_managers callback wins, the cache will miss the tools that had been removed from the cache
        # and will be blind to further changes in these tools.

        def reload_toolbox():
            save_integrated_tool_panel = False
            try:
                # Run and wait for toolbox reload on the process that watches the config files.
                # The toolbox reload will update the integrated_tool_panel_file
                self.app.queue_worker.send_local_control_task("reload_toolbox", get_response=True),
            except Exception:
                save_integrated_tool_panel = True
                log.exception("Exception occured while reloading toolbox")
            self.app.queue_worker.send_control_task(
                "reload_toolbox", noop_self=True, kwargs={"save_integrated_tool_panel": save_integrated_tool_panel}
            ),

        self.tool_config_watcher = get_tool_conf_watcher(
            reload_callback=reload_toolbox,
            tool_cache=self.app.tool_cache,
        )
        self.data_manager_config_watcher = get_tool_conf_watcher(
            reload_callback=lambda: self.app.queue_worker.send_control_task("reload_data_managers"),
        )
        self.tool_data_watcher = get_watcher(self.app.config, "watch_tool_data_dir", monitor_what_str="data tables")
        self.tool_watcher = get_tool_watcher(self, app.config)
        if getattr(self.app, "is_job_handler", False):
            self.job_rule_watcher = get_watcher(app.config, "watch_job_rules", monitor_what_str="job rules")
        else:
            self.job_rule_watcher = get_watcher(app.config, "__invalid__")
        self.core_config_watcher = get_watcher(app.config, "watch_core_config", monitor_what_str="core config file")
        self.tour_watcher = get_watcher(app.config, "watch_tours", monitor_what_str="tours")

    @property
    def watchers(self):
        return (
            self.tool_watcher,
            self.tool_config_watcher,
            self.data_manager_config_watcher,
            self.tool_data_watcher,
            self.tool_watcher,
            self.job_rule_watcher,
            self.core_config_watcher,
            self.tour_watcher,
        )

    def change_state(self, active):
        if active:
            self.start()
        elif self.active:
            self.shutdown()

    def start(self):
        for watcher in self.watchers:
            watcher.start()
        [self.tool_config_watcher.watch_file(config) for config in self.tool_config_paths]
        [self.data_manager_config_watcher.watch_file(config) for config in self.data_manager_configs]
        for tool_data_path in self.tool_data_paths:
            self.tool_data_watcher.watch_directory(
                tool_data_path,
                callback=lambda path: self.app.queue_worker.send_control_task(
                    "reload_tool_data_tables", kwargs={"path": path}
                ),
                require_extensions=(".loc",),
                recursive=True,
            )
        for job_rules_directory in self.job_rules_paths:
            self.job_rule_watcher.watch_directory(
                job_rules_directory,
                callback=lambda: self.app.queue_worker.send_control_task("reload_job_rules"),
                recursive=True,
                ignore_extensions=(".pyc", ".pyo", ".pyd"),
            )
        if self.app.config.config_file:
            self.core_config_watcher.watch_file(
                self.app.config.config_file,
                callback=lambda path: self.app.queue_worker.send_control_task("reload_core_config"),
            )
            self.tour_watcher.watch_directory(
                self.app.config.tour_config_dir,
                callback=lambda path: self.app.queue_worker.send_control_task("reload_tour", kwargs={"path": path}),
            )
        self.active = True

    def shutdown(self):
        for watcher in self.watchers:
            watcher.shutdown()
        self.active = False

    def update_watch_data_table_paths(self):
        if hasattr(self.tool_data_watcher, "monitored_dirs"):
            for tool_data_table_path in self.tool_data_paths:
                if tool_data_table_path not in self.tool_data_watcher.monitored_dirs:
                    self.tool_data_watcher.watch_directory(tool_data_table_path)

    @property
    def data_manager_configs(self):
        data_manager_configs = []
        if hasattr(self.app.config, "data_manager_config_file"):
            data_manager_configs.append(self.app.config.data_manager_config_file)
        if hasattr(self.app.config, "shed_data_manager_config_file"):
            data_manager_configs.append(self.app.config.shed_data_manager_config_file)
        return data_manager_configs

    @property
    def tool_data_paths(self):
        tool_data_paths = []
        if hasattr(self.app.config, "tool_data_path"):
            tool_data_paths.append(self.app.config.tool_data_path)
        if hasattr(self.app.config, "shed_tool_data_path"):
            tool_data_paths.append(self.app.config.shed_tool_data_path)
        return tool_data_paths

    @property
    def tool_config_paths(self):
        tool_config_paths = []
        if hasattr(self.app.config, "tool_configs"):
            tool_config_paths = self.app.config.tool_configs
        return tool_config_paths

    @property
    def job_rules_paths(self):
        job_rules_paths = []
        for rules_module in job_rule_modules(self.app):
            job_rules_dir = dirname(rules_module.__file__)
            job_rules_paths.append(job_rules_dir)
        return job_rules_paths



#!/usr/bin/env python
"""
Read a maf and split blocks by unique species combinations
"""
from __future__ import print_function

import sys

from bx.align import maf

from galaxy.tools.util import maf_utilities
from galaxy.util import string_as_bool


def __main__():
    try:
        maf_reader = maf.Reader(open(sys.argv[1]))
    except Exception as e:
        maf_utilities.tool_fail("Error opening MAF: %s" % e)
    try:
        out = maf.Writer(open(sys.argv[2], "w"))
    except Exception as e:
        maf_utilities.tool_fail("Error opening file for output: %s" % e)
    try:
        collapse_columns = string_as_bool(sys.argv[3])
    except Exception as e:
        maf_utilities.tool_fail("Error determining collapse columns value: %s" % e)

    start_count = 0
    end_count = 0
    for start_count, start_block in enumerate(maf_reader):  # noqa: B007
        for block in maf_utilities.iter_blocks_split_by_species(start_block):
            if collapse_columns:
                block.remove_all_gap_columns()
            out.write(block)
            end_count += 1
    out.close()

    if end_count:
        print("%i alignment blocks created from %i original blocks." % (end_count, start_count + 1))
    else:
        print("No alignment blocks were created.")


if __name__ == "__main__":
    __main__()



import errno
import logging
from typing import (
    Any,
    Dict,
    List,
)

from galaxy.exceptions import RequestParameterInvalidException
from galaxy.structured_app import MinimalManagerApp
from galaxy.tool_shed.util.basic_util import strip_path
from galaxy.tool_shed.util.repository_util import get_repository_owner
from galaxy.tool_shed.util.shed_util_common import get_tool_panel_config_tool_path_install_dir
from galaxy.util import (
    etree,
    parse_xml_string,
    xml_to_string,
)
from galaxy.util.renamed_temporary_file import RenamedTemporaryFile
from galaxy.util.tool_shed.common_util import remove_protocol_and_user_from_clone_url
from galaxy.util.tool_shed.xml_util import parse_xml

log = logging.getLogger(__name__)


class ToolPanelManager:
    app: MinimalManagerApp

    def __init__(self, app: MinimalManagerApp):
        self.app = app

    def add_to_shed_tool_config(self, shed_tool_conf_dict: Dict[str, Any], elem_list: list) -> None:
        """
        "A tool shed repository is being installed so change the shed_tool_conf file.  Parse the
        config file to generate the entire list of config_elems instead of using the in-memory list
        since it will be a subset of the entire list if one or more repositories have been deactivated.
        """
        if not elem_list:
            # We may have an empty elem_list in case a data manager is being installed.
            # In that case we don't want to wait for a toolbox reload that will never happen.
            return
        old_toolbox = self.app.toolbox
        shed_tool_conf = shed_tool_conf_dict["config_filename"]
        tool_cache_data_dir = shed_tool_conf_dict.get("tool_cache_data_dir")
        tool_path = shed_tool_conf_dict["tool_path"]
        config_elems = []
        # Ideally shed_tool_conf.xml would be created before the repo is cloned and added to the DB, but this is called
        # from too many places to make it feasible at this time
        try:
            tree, error_message = parse_xml(shed_tool_conf, check_exists=False)
        except OSError as exc:
            if exc.errno == errno.ENOENT and shed_tool_conf_dict.get("create", None) is not None:
                log.info("Creating shed tool config with default contents: %s", shed_tool_conf)
                with open(shed_tool_conf, "w") as fh:
                    fh.write(shed_tool_conf_dict["create"])
                tree, error_message = parse_xml(shed_tool_conf)
            else:
                log.error("Unable to load shed tool config: %s", shed_tool_conf)
                raise
        if tree:
            root = tree.getroot()
            for elem in root:
                config_elems.append(elem)
            # Add the new elements to the in-memory list of config_elems.
            for elem_entry in elem_list:
                if elem_entry.tag == "section":
                    # Loop through section entries in the in-memory tool panel.
                    for existing_elem in config_elems:
                        # Compare the section ID for each one to the section ID for the tool being installed.
                        if existing_elem.tag == "section" and existing_elem.attrib.get(
                            "id", None
                        ) == elem_entry.attrib.get("id", None):
                            # We've found a section, add the incoming tools to it.
                            for child in elem_entry:
                                existing_elem.append(child)
                            # Break out of the config_elems loop back to the elem_list loop.
                            break
                    # If we reach this point, no section was found. Create one with contents.
                    else:
                        config_elems.append(elem_entry)
                # This is not a section, but a tool or label. No need to search for matching sections, just add it.
                else:
                    config_elems.append(elem_entry)
            # Persist the altered shed_tool_config file.
            self.config_elems_to_xml_file(config_elems, shed_tool_conf, tool_path, tool_cache_data_dir)
            self.app.wait_for_toolbox_reload(old_toolbox)
        else:
            log.error(error_message)

    def add_to_tool_panel(
        self,
        repository_name,
        repository_clone_url,
        changeset_revision,
        repository_tools_tups,
        owner,
        shed_tool_conf,
        tool_panel_dict,
        new_install=True,
        tool_panel_section_mapping=None,
    ):
        """A tool shed repository is being installed or updated so handle tool panel alterations accordingly."""
        tool_panel_section_mapping = tool_panel_section_mapping or {}
        # We need to change the in-memory version and the file system version of the shed_tool_conf file.
        shed_tool_conf_dict = self.get_shed_tool_conf_dict(shed_tool_conf)
        tool_panel_dict = self.update_tool_panel_dict(
            tool_panel_dict, tool_panel_section_mapping, repository_tools_tups
        )
        # Generate the list of ElementTree Element objects for each section or tool.
        elem_list = self.generate_tool_panel_elem_list(
            repository_name,
            repository_clone_url,
            changeset_revision,
            tool_panel_dict,
            repository_tools_tups,
            owner=owner,
        )
        if new_install:
            tool_path = shed_tool_conf_dict["tool_path"]
            # Add the new elements to the shed_tool_conf file on disk.
            config_elems = shed_tool_conf_dict["config_elems"]
            for config_elem in elem_list:
                # Add the new elements to the in-memory list of config_elems.
                config_elems.append(config_elem)
                # Load the tools into the in-memory tool panel.
                self.app.toolbox.load_item(
                    config_elem,
                    tool_path=tool_path,
                    load_panel_dict=True,
                    guid=config_elem.get("guid"),
                )
            # Replace the old list of in-memory config_elems with the new list for this shed_tool_conf_dict.
            shed_tool_conf_dict["config_elems"] = config_elems
            self.app.toolbox.update_shed_config(shed_tool_conf_dict)
            self.add_to_shed_tool_config(shed_tool_conf_dict, elem_list)

    def config_elems_to_xml_file(self, config_elems, config_filename, tool_path, tool_cache_data_dir=None):
        """
        Persist the current in-memory list of config_elems to a file named by the
        value of config_filename.
        """
        try:
            tool_cache_data_dir = f' tool_cache_data_dir="{tool_cache_data_dir}"' if tool_cache_data_dir else ""
            root = parse_xml_string(
                f'<?xml version="1.0"?>\n<toolbox tool_path="{tool_path}"{tool_cache_data_dir}></toolbox>'
            )
            for elem in config_elems:
                root.append(elem)
            with RenamedTemporaryFile(config_filename, mode="w") as fh:
                fh.write(xml_to_string(root, pretty=True))
        except Exception:
            log.exception("Exception in ToolPanelManager.config_elems_to_xml_file")

    def generate_tool_elem(
        self, tool_shed, repository_name, changeset_revision, owner, tool_file_path, tool, tool_section
    ):
        """Create and return an ElementTree tool Element."""
        if tool_section is not None:
            tool_elem = etree.SubElement(tool_section, "tool")
        else:
            tool_elem = etree.Element("tool")
        tool_elem.attrib["file"] = tool_file_path
        if not tool.guid:
            raise ValueError("tool has no guid")
        tool_elem.attrib["guid"] = tool.guid
        tool_shed_elem = etree.SubElement(tool_elem, "tool_shed")
        tool_shed_elem.text = tool_shed
        repository_name_elem = etree.SubElement(tool_elem, "repository_name")
        repository_name_elem.text = repository_name
        repository_owner_elem = etree.SubElement(tool_elem, "repository_owner")
        repository_owner_elem.text = owner
        changeset_revision_elem = etree.SubElement(tool_elem, "installed_changeset_revision")
        changeset_revision_elem.text = changeset_revision
        id_elem = etree.SubElement(tool_elem, "id")
        id_elem.text = tool.id
        version_elem = etree.SubElement(tool_elem, "version")
        version_elem.text = tool.version
        return tool_elem

    def generate_tool_panel_dict_for_new_install(self, tool_dicts, tool_section=None):
        """
        When installing a repository that contains tools, all tools must
        currently be defined within the same tool section in the tool panel or
        outside of any sections.
        """
        tool_panel_dict: Dict[str, List[Dict[str, Any]]] = {}
        if tool_section:
            section_id = tool_section.id
            section_name = tool_section.name
            section_version = tool_section.version or ""
        else:
            section_id = ""
            section_name = ""
            section_version = ""
        for tool_dict in tool_dicts:
            if tool_dict.get("add_to_tool_panel", True):
                guid = tool_dict["guid"]
                tool_config = tool_dict["tool_config"]
                tool_section_dict = dict(
                    tool_config=tool_config, id=section_id, name=section_name, version=section_version
                )
                if guid in tool_panel_dict:
                    tool_panel_dict[guid].append(tool_section_dict)
                else:
                    tool_panel_dict[guid] = [tool_section_dict]
        return tool_panel_dict

    def generate_tool_panel_dict_for_tool_config(
        self, guid, tool_config, tool_sections=None
    ) -> Dict[str, List[Dict[str, Any]]]:
        """
        Create a dictionary of the following type for a single tool config file name.
        The intent is to call this method for every tool config in a repository and
        append each of these as entries to a tool panel dictionary for the repository.
        This enables each tool to be loaded into a different section in the tool panel.

        .. code-block::

            {<Tool guid> :
                [{ tool_config : <tool_config_file>,
                    id: <ToolSection id>,
                    version : <ToolSection version>,
                    name : <TooSection name>}]}

        """
        tool_panel_dict: Dict[str, List[Dict[str, Any]]] = {}
        file_name = strip_path(tool_config)
        tool_section_dicts = self.generate_tool_section_dicts(tool_config=file_name, tool_sections=tool_sections)
        tool_panel_dict[guid] = tool_section_dicts
        return tool_panel_dict

    def generate_tool_panel_dict_from_shed_tool_conf_entries(self, repository) -> Dict[str, List[Dict[str, Any]]]:
        """
        Keep track of the section in the tool panel in which this repository's
        tools will be contained by parsing the shed_tool_conf in which the
        repository's tools are defined and storing the tool panel definition
        of each tool in the repository. This method is called only when the
        repository is being deactivated or un-installed and allows for
        activation or re-installation using the original layout.
        """
        tool_panel_dict: Dict[str, List[Dict[str, Any]]] = {}
        shed_tool_conf, tool_path, relative_install_dir = get_tool_panel_config_tool_path_install_dir(
            self.app, repository
        )
        metadata = repository.metadata_
        # Create a dictionary of tool guid and tool config file name for each tool in the repository.
        guids_and_configs = {}
        if "tools" in metadata:
            for tool_dict in metadata["tools"]:
                guid = tool_dict["guid"]
                tool_config = tool_dict["tool_config"]
                file_name = strip_path(tool_config)
                guids_and_configs[guid] = file_name
        # Parse the shed_tool_conf file in which all of this repository's tools are defined and generate the tool_panel_dict.
        tree, error_message = parse_xml(shed_tool_conf)
        if tree is None:
            return tool_panel_dict
        root = tree.getroot()
        for elem in root:
            if elem.tag == "tool":
                guid = elem.get("guid")
                if guid in guids_and_configs:
                    # The tool is displayed in the tool panel outside of any tool sections.
                    tool_section_dict = dict(tool_config=guids_and_configs[guid], id="", name="", version="")
                    if guid in tool_panel_dict:
                        tool_panel_dict[guid].append(tool_section_dict)
                    else:
                        tool_panel_dict[guid] = [tool_section_dict]
            elif elem.tag == "section":
                section_id = elem.get("id") or ""
                section_name = elem.get("name") or ""
                section_version = elem.get("version") or ""
                for section_elem in elem:
                    if section_elem.tag == "tool":
                        guid = section_elem.get("guid")
                        if guid in guids_and_configs:
                            # The tool is displayed in the tool panel inside the current tool section.
                            tool_section_dict = dict(
                                tool_config=guids_and_configs[guid],
                                id=section_id,
                                name=section_name,
                                version=section_version,
                            )
                            if guid in tool_panel_dict:
                                tool_panel_dict[guid].append(tool_section_dict)
                            else:
                                tool_panel_dict[guid] = [tool_section_dict]
        return tool_panel_dict

    def generate_tool_panel_elem_list(
        self,
        repository_name: str,
        repository_clone_url: str,
        changeset_revision: str,
        tool_panel_dict: dict,
        repository_tools_tups: List[tuple],
        owner="",
    ):
        """Generate a list of ElementTree Element objects for each section or tool."""
        elem_list: List[etree.Element] = []
        tool_elem = None
        cleaned_repository_clone_url = remove_protocol_and_user_from_clone_url(repository_clone_url)
        if not owner:
            owner = get_repository_owner(cleaned_repository_clone_url)
        tool_shed = cleaned_repository_clone_url.split("/repos/")[0].rstrip("/")
        for guid, tool_section_dicts in tool_panel_dict.items():
            for tool_section_dict in tool_section_dicts:
                tool_section = None
                inside_section = False
                section_in_elem_list = None
                if tool_section_dict["id"]:
                    inside_section = True
                    # Create a new section element only if we haven't already created it.
                    for index, elem in enumerate(elem_list):
                        if elem.tag == "section":
                            section_id = elem.get("id", None)
                            if section_id == tool_section_dict["id"]:
                                section_in_elem_list = index
                                tool_section = elem
                                break
                    if tool_section is None:
                        tool_section = self.generate_tool_section_element_from_dict(tool_section_dict)
                # Find the tuple containing the current guid from the list of repository_tools_tups.
                for repository_tool_tup in repository_tools_tups:
                    tool_file_path, tup_guid, tool = repository_tool_tup
                    if tup_guid == guid:
                        break
                tool_elem = self.generate_tool_elem(
                    tool_shed,
                    repository_name,
                    changeset_revision,
                    owner,
                    tool_file_path,
                    tool,
                    tool_section if inside_section else None,
                )
                if inside_section:
                    if section_in_elem_list is not None:
                        elem_list[section_in_elem_list] = tool_section
                    else:
                        elem_list.append(tool_section)
                else:
                    elem_list.append(tool_elem)
        return elem_list

    def generate_tool_section_dicts(self, tool_config=None, tool_sections=None) -> List[Dict[str, Any]]:
        tool_section_dicts: List[Dict[str, Any]] = []
        if tool_config is None:
            tool_config = ""
        if tool_sections:
            for tool_section in tool_sections:
                # The value of tool_section will be None if the tool is displayed outside
                # of any sections in the tool panel.
                if tool_section:
                    section_id = tool_section.id or ""
                    section_version = tool_section.version or ""
                    section_name = tool_section.name or ""
                else:
                    section_id = ""
                    section_version = ""
                    section_name = ""
                tool_section_dicts.append(
                    dict(tool_config=tool_config, id=section_id, version=section_version, name=section_name)
                )
        else:
            tool_section_dicts.append(dict(tool_config=tool_config, id="", version="", name=""))
        return tool_section_dicts

    def generate_tool_section_element_from_dict(self, tool_section_dict):
        # The value of tool_section_dict looks like the following.
        # { id: <ToolSection id>, version : <ToolSection version>, name : <TooSection name>}
        if tool_section_dict["id"]:
            # Create a new tool section.
            tool_section = etree.Element("section")
            tool_section.attrib["id"] = tool_section_dict["id"]
            tool_section.attrib["name"] = tool_section_dict["name"]
            tool_section.attrib["version"] = tool_section_dict["version"]
        else:
            tool_section = None
        return tool_section

    def get_or_create_tool_section(self, toolbox, tool_panel_section_id, new_tool_panel_section_label=None):
        return toolbox.get_section(
            section_id=tool_panel_section_id, new_label=new_tool_panel_section_label, create_if_needed=True
        )

    def get_shed_tool_conf_dict(self, shed_tool_conf):
        """
        Return the in-memory version of the shed_tool_conf file, which is stored in
        the config_elems entry in the shed_tool_conf_dict associated with the file.
        """
        for shed_tool_conf_dict in self.app.toolbox.dynamic_confs(include_migrated_tool_conf=True):
            if shed_tool_conf == shed_tool_conf_dict["config_filename"]:
                return shed_tool_conf_dict
            else:
                file_name = strip_path(shed_tool_conf_dict["config_filename"])
                if shed_tool_conf == file_name:
                    return shed_tool_conf_dict
        raise RequestParameterInvalidException(
            f"Requested shed_tool_conf '{shed_tool_conf}' is not an active shed_tool_config_file"
        )

    def handle_tool_panel_section(self, toolbox, tool_panel_section_id=None, new_tool_panel_section_label=None):
        """Return a ToolSection object retrieved from the current in-memory tool_panel."""
        # If tool_panel_section_id is received, the section exists in the tool panel.  In this
        # case, the value of the received tool_panel_section_id must be the id retrieved from a
        # tool panel config (e.g., tool_conf.xml, which may have getext).  If new_tool_panel_section_label
        # is received, a new section will be added to the tool panel.
        if new_tool_panel_section_label:
            section_id = str(new_tool_panel_section_label.lower().replace(" ", "_"))
            tool_panel_section_key, tool_section = self.get_or_create_tool_section(
                toolbox, tool_panel_section_id=section_id, new_tool_panel_section_label=new_tool_panel_section_label
            )
        elif tool_panel_section_id:
            tool_panel_section_key, tool_section = toolbox.get_section(tool_panel_section_id)
        else:
            return None, None
        return tool_panel_section_key, tool_section

    def handle_tool_panel_selection(
        self, toolbox, metadata, no_changes_checked, tool_panel_section_id, new_tool_panel_section_label
    ):
        """
        Handle the selected tool panel location for loading tools included in
        tool shed repositories when installing or reinstalling them.
        """
        # Get the location in the tool panel in which each tool was originally loaded.
        tool_section = None
        tool_panel_section_key = None
        if "tools" in metadata:
            # This forces everything to be loaded into the same section (or no section)
            # in the tool panel.
            if no_changes_checked:
                # Make sure the no_changes check box overrides the new_tool_panel_section_label
                # if the user checked the check box and entered something into the field.
                new_tool_panel_section_label = None
                if "tool_panel_section" in metadata:
                    tool_panel_dict = metadata["tool_panel_section"]
                    if not tool_panel_dict:
                        tool_panel_dict = self.generate_tool_panel_dict_for_new_install(metadata["tools"])
                else:
                    tool_panel_dict = self.generate_tool_panel_dict_for_new_install(metadata["tools"])
                if tool_panel_dict:
                    # The tool_panel_dict is empty when tools exist but are not installed into a tool panel section.
                    tool_section_dicts = tool_panel_dict[next(iter(tool_panel_dict))]
                    tool_section_dict = tool_section_dicts[0]
                    original_section_id = tool_section_dict["id"]
                    if original_section_id:
                        tool_panel_section_key, tool_section = self.get_or_create_tool_section(
                            toolbox,
                            tool_panel_section_id=original_section_id,
                            new_tool_panel_section_label=new_tool_panel_section_label,
                        )
            else:
                # The user elected to change the tool panel section to contain the tools.
                tool_panel_section_key, tool_section = self.handle_tool_panel_section(
                    toolbox,
                    tool_panel_section_id=tool_panel_section_id,
                    new_tool_panel_section_label=new_tool_panel_section_label,
                )
        return tool_section, tool_panel_section_key

    def remove_from_shed_tool_config(self, shed_tool_conf_dict, metadata):
        """
        A tool shed repository is being uninstalled so change the
        shed_tool_conf file. Parse the config file to generate the entire list
        of config_elems instead of using the in-memory list since it will be a
        subset of the entire list if one or more repositories have been
        deactivated.
        """
        if "tools" not in metadata:
            return
        # We need to use the tool path to uniquely identify the tools to remove,
        # since multiple installable revisions of a repository can provide the
        # same version of a tool (i.e. there may be another tool with the same
        # guid that we should not remove from shed_tool_conf).
        guid_paths_to_remove = [(_["guid"], _["tool_config"]) for _ in metadata["tools"]]
        shed_tool_conf = shed_tool_conf_dict["config_filename"]
        tool_path = shed_tool_conf_dict["tool_path"]
        config_elems = []
        tree, error_message = parse_xml(shed_tool_conf)
        if tree:
            root = tree.getroot()
            for elem in root:
                config_elems.append(elem)
            config_elems_to_remove = []
            for config_elem in config_elems:
                if config_elem.tag == "section":
                    tool_elems_to_remove = []
                    for tool_elem in config_elem:
                        if (tool_elem.get("guid"), tool_elem.get("file")) in guid_paths_to_remove:
                            tool_elems_to_remove.append(tool_elem)
                    for tool_elem in tool_elems_to_remove:
                        # Remove all of the appropriate tool sub-elements from the section element.
                        config_elem.remove(tool_elem)
                    if len(config_elem) < 1:
                        # Keep a list of all empty section elements so they can be removed.
                        config_elems_to_remove.append(config_elem)
                elif config_elem.tag == "tool":
                    if (config_elem.get("guid"), config_elem.get("file")) in guid_paths_to_remove:
                        config_elems_to_remove.append(config_elem)
            for config_elem in config_elems_to_remove:
                config_elems.remove(config_elem)
            # Persist the altered in-memory version of the tool config.
            self.config_elems_to_xml_file(config_elems, shed_tool_conf, tool_path)

    def remove_repository_contents(self, repository, shed_tool_conf, uninstall):
        """
        A tool shed repository is being deactivated or uninstalled, so handle
        tool panel alterations accordingly.
        """
        # Determine where the tools are currently defined in the tool panel and store this
        # information so the tools can be displayed in the same way when the repository is
        # activated or reinstalled.
        tool_panel_dict = self.generate_tool_panel_dict_from_shed_tool_conf_entries(repository)
        repository.metadata_["tool_panel_section"] = tool_panel_dict
        self.app.install_model.context.add(repository)
        self.app.install_model.context.flush()
        # Create a list of guids for all tools that will be removed from the in-memory tool panel
        # and config file on disk.
        guids_to_remove = list(tool_panel_dict.keys())
        toolbox = self.app.toolbox
        # Remove the tools from the toolbox's tools_by_id dictionary.
        for guid_to_remove in guids_to_remove:
            # remove_from_tool_panel to false, will handling that logic below.
            toolbox.remove_tool_by_id(guid_to_remove, remove_from_panel=False)
        shed_tool_conf_dict = self.get_shed_tool_conf_dict(shed_tool_conf)
        if uninstall:
            # Remove from the shed_tool_conf file on disk.
            self.remove_from_shed_tool_config(shed_tool_conf_dict, repository.metadata_)

    def update_tool_panel_dict(self, tool_panel_dict, tool_panel_section_mapping, repository_tools_tups):
        for tool_guid in tool_panel_dict:
            if tool_guid not in tool_panel_section_mapping:
                continue
            for tool in tool_panel_dict[tool_guid]:
                section_name = tool_panel_section_mapping[tool_guid]["tool_panel_section"]
                section_id = str(tool_panel_section_mapping[tool_guid]["tool_panel_section"].lower().replace(" ", "_"))
                tool["name"] = section_name
                tool["id"] = section_id
        return tool_panel_dict



#!/usr/bin/env python

"""
Read a maf and output intervals for specified list of species.
"""
import os
import sys

from bx.align import maf

from galaxy.tools.util import maf_utilities


def __main__():
    input_filename = sys.argv[1]
    output_filename = sys.argv[2]
    output_id = sys.argv[3]
    # where to store files that become additional output
    database_tmp_dir = sys.argv[4]
    primary_spec = sys.argv[5]
    species = sys.argv[6].split(",")
    all_species = sys.argv[7].split(",")
    partial = sys.argv[8]
    keep_gaps = sys.argv[9]
    out_files = {}

    if "None" in species:
        species = []

    if primary_spec not in species:
        species.append(primary_spec)
    if primary_spec not in all_species:
        all_species.append(primary_spec)

    all_species.sort()
    for spec in species:
        if spec == primary_spec:
            out_files[spec] = open(output_filename, "w+")
        else:
            out_files[spec] = open(
                os.path.join(database_tmp_dir, "primary_%s_%s_visible_interval_%s" % (output_id, spec, spec)), "w+"
            )
        out_files[spec].write("#chrom\tstart\tend\tstrand\tscore\tname\t%s\n" % ("\t".join(all_species)))
    num_species = len(all_species)

    file_in = open(input_filename)
    maf_reader = maf.Reader(file_in)

    for i, m in enumerate(maf_reader):
        for j, block in enumerate(maf_utilities.iter_blocks_split_by_species(m)):
            if len(block.components) < num_species and partial == "partial_disallowed":
                continue
            sequences = {}
            for c in block.components:
                spec, chrom = maf_utilities.src_split(c.src)
                if keep_gaps == "remove_gaps":
                    sequences[spec] = c.text.replace("-", "")
                else:
                    sequences[spec] = c.text
            sequences = "\t".join(sequences.get(_, "") for _ in all_species)
            for spec in species:
                c = block.get_component_by_src_start(spec)
                if c is not None:
                    spec2, chrom = maf_utilities.src_split(c.src)
                    assert spec2 == spec, Exception(
                        "Species name inconsistancy found in component: %s != %s" % (spec, spec2)
                    )
                    out_files[spec].write(
                        "%s\t%s\t%s\t%s\t%s\t%s\t%s\n"
                        % (
                            chrom,
                            c.forward_strand_start,
                            c.forward_strand_end,
                            c.strand,
                            m.score,
                            "%s_%s_%s" % (spec, i, j),
                            sequences,
                        )
                    )
    file_in.close()
    for file_out in out_files.values():
        file_out.close()


if __name__ == "__main__":
    __main__()



#!/usr/bin/env python
# This tool takes a gff file as input and creates filters on attributes based on certain properties.
# The tool will skip over invalid lines within the file, informing the user about the number of lines skipped.
# TODO: much of this code is copied from the Filter1 tool (filtering.py in tools/stats/). The commonalities should be
# abstracted and leveraged in each filtering tool.
from __future__ import (
    division,
    print_function,
)

import sys
from ast import (
    Module,
    parse,
    walk,
)
from json import loads

AST_NODE_TYPE_WHITELIST = [
    "Expr",
    "Load",
    "Str",
    "Num",
    "BoolOp",
    "Compare",
    "And",
    "Eq",
    "NotEq",
    "Or",
    "GtE",
    "LtE",
    "Lt",
    "Gt",
    "BinOp",
    "Add",
    "Div",
    "Sub",
    "Mult",
    "Mod",
    "Pow",
    "LShift",
    "GShift",
    "BitAnd",
    "BitOr",
    "BitXor",
    "UnaryOp",
    "Invert",
    "Not",
    "NotIn",
    "In",
    "Is",
    "IsNot",
    "List",
    "Index",
    "Subscript",
    "Name",
]


BUILTIN_AND_MATH_FUNCTIONS = "abs|all|any|bin|chr|cmp|complex|divmod|float|hex|int|len|long|max|min|oct|ord|pow|range|reversed|round|sorted|str|sum|type|unichr|unicode|log|exp|sqrt|ceil|floor".split(
    "|"
)
STRING_AND_LIST_METHODS = [name for name in dir("") + dir([]) if not name.startswith("_")]
VALID_FUNCTIONS = BUILTIN_AND_MATH_FUNCTIONS + STRING_AND_LIST_METHODS
# Name blacklist isn't strictly needed - but provides extra peace of mind.
NAME_BLACKLIST = ["exec", "eval", "globals", "locals", "__import__", "__builtins__"]


def __check_name(ast_node):
    name = ast_node.id
    return name not in NAME_BLACKLIST


def check_simple_name(text):
    """

    >>> check_simple_name("col_name")
    True
    >>> check_simple_name("c1=='chr1' and c3-c2>=2000 and c6=='+'")
    False
    >>> check_simple_name("eval('1+1')")
    False
    >>> check_simple_name("import sys")
    False
    >>> check_simple_name("[].__str__")
    False
    >>> check_simple_name("__builtins__")
    False
    >>> check_simple_name("'x' in globals")
    False
    >>> check_simple_name("'x' in [1,2,3]")
    False
    >>> check_simple_name("c3=='chr1' and c5>5")
    False
    >>> check_simple_name("c3=='chr1' and d5>5")
    False
    >>> check_simple_name("c3=='chr1' and c5>5 or exec")
    False
    >>> check_simple_name("type(c1) != type(1)")
    False
    >>> check_simple_name("c1.split(',')[1] == '1'")
    False
    >>> check_simple_name("exec 1")
    False
    >>> check_simple_name("str(c2) in [\\\"a\\\",\\\"b\\\"]")
    False
    >>> check_simple_name("__import__('os').system('touch /tmp/OOPS')")
    False
    """
    try:
        module = parse(text)
    except SyntaxError:
        return False

    if not isinstance(module, Module):
        return False
    statements = module.body
    if not len(statements) == 1:
        return False
    expression = statements[0]
    if expression.__class__.__name__ != "Expr":
        return False

    for ast_node in walk(expression):
        ast_node_class = ast_node.__class__.__name__
        if ast_node_class not in ["Expr", "Name", "Load"]:
            return False

        if ast_node_class == "Name" and not __check_name(ast_node):
            return False

    return True


def check_expression(text):
    """

    >>> check_expression("c1=='chr1' and c3-c2>=2000 and c6=='+'")
    True
    >>> check_expression("eval('1+1')")
    False
    >>> check_expression("import sys")
    False
    >>> check_expression("[].__str__")
    False
    >>> check_expression("__builtins__")
    False
    >>> check_expression("'x' in globals")
    False
    >>> check_expression("'x' in [1,2,3]")
    True
    >>> check_expression("c3=='chr1' and c5>5")
    True
    >>> check_expression("c3=='chr1' and d5>5")
    True
    >>> check_expression("c3=='chr1' and c5>5 or exec")
    False
    >>> check_expression("type(c1) != type(1)")
    False
    >>> check_expression("c1.split(',')[1] == '1'")
    False
    >>> check_expression("exec 1")
    False
    >>> check_expression("str(c2) in [\\\"a\\\",\\\"b\\\"]")
    False
    >>> check_expression("__import__('os').system('touch /tmp/OOPS')")
    False
    """
    try:
        module = parse(text)
    except SyntaxError:
        return False

    if not isinstance(module, Module):
        return False
    statements = module.body
    if not len(statements) == 1:
        return False
    expression = statements[0]
    if expression.__class__.__name__ != "Expr":
        return False

    for ast_node in walk(expression):
        ast_node_class = ast_node.__class__.__name__

        # Toss out everything that is not a "simple" expression,
        # imports, error handling, etc...
        if ast_node_class not in AST_NODE_TYPE_WHITELIST:
            return False

        if ast_node_class == "Name" and not __check_name(ast_node):
            return False

    return True


#
# Helper functions.
#
def get_operands(filter_condition):
    # Note that the order of all_operators is important
    items_to_strip = [
        "+",
        "-",
        "**",
        "*",
        "//",
        "/",
        "%",
        "<<",
        ">>",
        "&",
        "|",
        "^",
        "~",
        "<=",
        "<",
        ">=",
        ">",
        "==",
        "!=",
        "<>",
        " and ",
        " or ",
        " not ",
        " is ",
        " is not ",
        " in ",
        " not in ",
    ]
    for item in items_to_strip:
        if filter_condition.find(item) >= 0:
            filter_condition = filter_condition.replace(item, " ")
    operands = set(filter_condition.split(" "))
    return operands


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit()


def check_for_executable(text, description=""):
    # Attempt to determine if the condition includes executable stuff and, if so, exit.
    secured = dir()
    operands = get_operands(text)
    for operand in operands:
        try:
            int(operand)
        except ValueError:
            if operand in secured:
                stop_err("Illegal value '%s' in %s '%s'" % (operand, description, text))


#
# Process inputs.
#
in_fname = sys.argv[1]
out_fname = sys.argv[2]
cond_text = sys.argv[3]
attribute_types = loads(sys.argv[4])

# Convert types from str to type objects.
for name, a_type in attribute_types.items():
    check_for_executable(a_type)
    if not check_simple_name(a_type):
        stop_err("Problem with attribute type [%s]" % a_type)
    attribute_types[name] = eval(a_type)

# Possible (eg workflows) that user's filter contains standard
# GFF attributes which are not present in the actual file -
# and thus not in the metadata passed in by Galaxy.
# To avoid a nasty error here, add the official terms from
# the GFF3 specification (if not already defined).
# (These all start with a capital letter, which is important):
for name in [
    "ID",
    "Name",
    "Alias",
    "Parent",
    "Target",
    "Gap",
    "Derives_from",
    "Note",
    "Dbxref",
    "Ontology_term",
    "Is_circular",
]:
    attribute_types[name] = str

# Unescape if input has been escaped
mapped_str = {
    "__lt__": "<",
    "__le__": "<=",
    "__eq__": "==",
    "__ne__": "!=",
    "__gt__": ">",
    "__ge__": ">=",
    "__sq__": "'",
    "__dq__": '"',
}
for key, value in mapped_str.items():
    cond_text = cond_text.replace(key, value)

# Attempt to determine if the condition includes executable stuff and, if so, exit.
check_for_executable(cond_text, "condition")

if not check_expression(cond_text):
    stop_err("Illegal/invalid in condition '%s'" % (cond_text))

# Prepare the column variable names and wrappers for column data types. Only
# prepare columns up to largest column in condition.
attrs, type_casts = [], []
for name in attribute_types.keys():
    attrs.append(name)
    type_cast = "get_value('%(name)s', attribute_types['%(name)s'], attribute_values)" % ({"name": name})
    type_casts.append(type_cast)

attr_str = ", ".join(attrs)  # 'c1, c2, c3, c4'
type_cast_str = ", ".join(type_casts)  # 'str(c1), int(c2), int(c3), str(c4)'
wrap = "%s = %s" % (attr_str, type_cast_str)

# Stats
skipped_lines = 0
first_invalid_line = 0
invalid_line = None
lines_kept = 0
total_lines = 0
out = open(out_fname, "wt")


# Helper function to safely get and type cast a value in a dict.
def get_value(name, a_type, values_dict):
    if name in values_dict:
        return (a_type)(values_dict[name])
    else:
        return None


# Read and filter input file, skipping invalid lines
code = """
for i, line in enumerate( open( in_fname ) ):
    total_lines += 1
    line = line.rstrip( '\\r\\n' )
    if not line or line.startswith( '#' ):
        # Ignore blank lines or comments
        continue
    try:
        # Place attribute values into variables with attribute
        # name; type casting is done as well.
        elems = line.split( '\t' )
        attribute_values = {}
        for name_value_pair in elems[8].split(";"):
            # Split on first equals (GFF3) or space (legacy)
            name_value_pair = name_value_pair.strip()
            i = name_value_pair.replace(" ", "=").find("=")
            if i == -1:
                continue
            name = name_value_pair[:i].strip()
            if name == '':
                continue
            # Need to strip double quote from value and typecast.
            attribute_values[name] = name_value_pair[i+1:].strip(" \\"")
        %s
        if %s:
            lines_kept += 1
            print( line, file=out )
    except Exception as e:
        print( e )
        skipped_lines += 1
        if not invalid_line:
            first_invalid_line = i + 1
            invalid_line = line
""" % (
    wrap,
    cond_text,
)

valid_filter = True
try:
    exec(code)
except Exception as e:
    out.close()
    if str(e).startswith("invalid syntax"):
        valid_filter = False
        stop_err('Filter condition "%s" likely invalid. See tool tips, syntax and examples.' % cond_text)
    else:
        stop_err(str(e))

if valid_filter:
    out.close()
    valid_lines = total_lines - skipped_lines
    print("Filtering with %s, " % (cond_text))
    if valid_lines > 0:
        print("kept %4.2f%% of %d lines." % (100.0 * lines_kept / valid_lines, total_lines))
    else:
        print(
            'Possible invalid filter condition "%s" or non-existent column referenced. See tool tips, syntax and examples.'
            % cond_text
        )
    if skipped_lines > 0:
        print('Skipped %d invalid lines starting at line #%d: "%s"' % (skipped_lines, first_invalid_line, invalid_line))



import logging
from typing import (
    Optional,
    TYPE_CHECKING,
    Union,
)
from uuid import UUID

from sqlalchemy import sql

from galaxy import (
    exceptions,
    model,
)
from galaxy.exceptions import DuplicatedIdentifierException
from galaxy.tool_util.cwl import tool_proxy
from .base import (
    ModelManager,
    raise_filter_err,
)
from .executables import artifact_class

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from galaxy.managers.base import OrmFilterParsersType


class DynamicToolManager(ModelManager):
    """Manages dynamic tools stored in Galaxy's database."""

    model_class = model.DynamicTool

    def get_tool_by_uuid(self, uuid: Optional[Union[UUID, str]]):
        dynamic_tool = self._one_or_none(self.query().filter(self.model_class.uuid == uuid))
        return dynamic_tool

    def get_tool_by_tool_id(self, tool_id):
        dynamic_tool = self._one_or_none(self.query().filter(self.model_class.tool_id == tool_id))
        return dynamic_tool

    def get_tool_by_id(self, object_id):
        dynamic_tool = self._one_or_none(self.query().filter(self.model_class.id == object_id))
        return dynamic_tool

    def create_tool(self, trans, tool_payload, allow_load=True):
        if not getattr(self.app.config, "enable_beta_tool_formats", False):
            raise exceptions.ConfigDoesNotAllowException(
                "Set 'enable_beta_tool_formats' in Galaxy config to create dynamic tools."
            )

        dynamic_tool = None
        uuid_str = tool_payload.get("uuid")
        # Convert uuid_str to UUID or generate new if None
        uuid = model.get_uuid(uuid_str)
        if uuid_str:
            # TODO: enforce via DB constraint and catch appropriate
            # exception.
            dynamic_tool = self.get_tool_by_uuid(uuid_str)
            if dynamic_tool:
                if not allow_load:
                    raise DuplicatedIdentifierException(dynamic_tool.id)
                assert dynamic_tool.uuid == uuid
        if not dynamic_tool:
            src = tool_payload.get("src", "representation")
            is_path = src == "from_path"

            if is_path:
                tool_format, representation, _ = artifact_class(None, tool_payload)
            else:
                assert src == "representation"
                representation = tool_payload.get("representation")
                if not representation:
                    raise exceptions.ObjectAttributeMissingException("A tool 'representation' is required.")

                tool_format = representation.get("class")
                if not tool_format:
                    raise exceptions.ObjectAttributeMissingException("Current tool representations require 'class'.")

            tool_path = tool_payload.get("path")
            tool_directory = tool_payload.get("tool_directory")
            if tool_format == "GalaxyTool":
                tool_id = representation.get("id")
                if not tool_id:
                    tool_id = str(uuid)
            elif tool_format in ("CommandLineTool", "ExpressionTool"):
                # CWL tools
                if is_path:
                    proxy = tool_proxy(tool_path=tool_path, uuid=uuid)
                else:
                    # Build a tool proxy so that we can convert to the persistable
                    # hash.
                    proxy = tool_proxy(
                        tool_object=representation["raw_process_reference"],
                        tool_directory=tool_directory,
                        uuid=uuid,
                    )
                tool_id = proxy.galaxy_id()
            else:
                raise Exception(f"Unknown tool format [{tool_format}] encountered.")
            tool_version = representation.get("version")
            dynamic_tool = self.create(
                tool_format=tool_format,
                tool_id=tool_id,
                tool_version=tool_version,
                tool_path=tool_path,
                tool_directory=tool_directory,
                uuid=uuid,
                value=representation,
            )
        self.app.toolbox.load_dynamic_tool(dynamic_tool)
        return dynamic_tool

    def list_tools(self, active=True):
        return self.query().filter(self.model_class.active == active)

    def deactivate(self, dynamic_tool):
        self.update(dynamic_tool, {"active": False})
        return dynamic_tool


class ToolFilterMixin:
    orm_filter_parsers: "OrmFilterParsersType"

    def create_tool_filter(self, attr, op, val):
        def _create_tool_filter(model_class=None):
            if op == "eq":
                cond = model.Job.table.c.tool_id == val
            elif op == "contains":
                cond = model.Job.table.c.tool_id.contains(val, autoescape=True)
            else:
                raise_filter_err(attr, op, val, "bad op in filter")
            if model_class is model.HistoryDatasetAssociation:
                return sql.expression.and_(
                    model.Job.table.c.id == model.JobToOutputDatasetAssociation.table.c.job_id,
                    model.HistoryDatasetAssociation.table.c.id
                    == model.JobToOutputDatasetAssociation.table.c.dataset_id,
                    cond,
                )
            elif model_class is model.HistoryDatasetCollectionAssociation:
                return sql.expression.and_(
                    model.Job.id == model.JobToOutputDatasetAssociation.job_id,
                    model.JobToOutputDatasetAssociation.dataset_id == model.DatasetCollectionElement.hda_id,
                    model.DatasetCollectionElement.dataset_collection_id
                    == model.HistoryDatasetCollectionAssociation.collection_id,
                    cond,
                )
            else:
                return True

        return _create_tool_filter

    def _add_parsers(self):
        self.orm_filter_parsers.update(
            {
                "tool_id": self.create_tool_filter,
            }
        )



""" This module contains utility methods for reasoning about and ordering
workflow steps.
"""
import math

from galaxy.util.topsort import (
    CycleError,
    topsort,
    topsort_levels,
)


def attach_ordered_steps(workflow):
    """Attempt to topologically order steps and attach to workflow. If this
    fails - the workflow contains cycles so it mark it as such.
    """
    ordered_steps = order_workflow_steps(workflow.steps)
    workflow.has_cycles = True
    if ordered_steps:
        workflow.has_cycles = False
        workflow.steps = ordered_steps
    for i, step in enumerate(workflow.steps):
        step.order_index = i
    return workflow.has_cycles


def order_workflow_steps(steps):
    """
    Perform topological sort of the steps, return ordered or None
    """
    position_data_available = bool(steps)
    for step in steps:
        if step.subworkflow:
            attach_ordered_steps(step.subworkflow)
        if not step.position or "left" not in step.position or "top" not in step.position:
            position_data_available = False
    if position_data_available:
        # find minimum left and top values to normalize position
        min_left = min(step.position["left"] for step in steps)
        min_top = min(step.position["top"] for step in steps)
        # normalize by min_left and min_top
        for step in steps:
            step.position = {
                "left": step.position["left"] - min_left,
                "top": step.position["top"] - min_top
                # other position attributes can be discarded if present
            }
        steps.sort(key=lambda _: _.position["left"] + _.position["top"])
        # order by Euclidean distance to the origin (i.e. pre-normalized (min_left, min_top))
        steps.sort(key=lambda _: math.sqrt(_.position["left"] ** 2 + _.position["top"] ** 2))
    try:
        edges = sorted(edgelist_for_workflow_steps(steps))
        node_order = topsort(edges)
        return [steps[i] for i in node_order]
    except CycleError:
        return None


def has_cycles(workflow):
    try:
        topsort(sorted(edgelist_for_workflow_steps(workflow.steps)))
        return False
    except CycleError:
        return True


def edgelist_for_workflow_steps(steps):
    """
    Create a list of tuples representing edges between ``WorkflowStep`` s based
    on associated ``WorkflowStepConnection`` s
    """
    edges = []
    steps_to_index = {step: i for i, step in enumerate(steps)}
    for step in steps:
        edges.append((steps_to_index[step], steps_to_index[step]))
        for conn in step.input_connections:
            output_index = steps_to_index[conn.output_step]
            input_index = steps_to_index[conn.input_step]
            # self connection - a cycle not detectable by topsort function.
            if output_index == input_index:
                raise CycleError([], 0, 0)
            edges.append((output_index, input_index))
    return edges


def order_workflow_steps_with_levels(steps):
    try:
        return topsort_levels(edgelist_for_workflow_steps(steps))
    except CycleError:
        return None



import fcntl

from galaxy.util import unicodify


def fork_safe_write(path: str, contents: str):
    # The following write method looks a little funky and is inspired by https://twitter.com/_monoid/status/1317895053122150400.
    # This should guarantee that we wait until all forks that inherit open FDs have proceeded to execve,
    # because only then the shared lock can be acquired.
    # This **should** entirely avoid the "Text File Busy" error that `_handle_script_integrity` attempts to deal with.
    # The likelihood of "Text File Busy" happening increases as the load increases, and more work has to be done
    # when forking to copy memory pages, making the fork slower and therefore more likely to happen while the script
    # file is open for writing.
    with open(path, "w", encoding="utf-8") as f:
        f.write(unicodify(contents))
        fcntl.flock(f.fileno(), fcntl.LOCK_EX)
    with open(path) as f:
        fcntl.flock(f.fileno(), fcntl.LOCK_SH)



import json
import logging
from typing import List

from pydantic import (
    BaseModel,
    Field,
    HttpUrl,
)

from galaxy import exceptions
from galaxy.util.resources import resource_string

log = logging.getLogger(__name__)


# https://github.com/spdx/license-list-data/blob/master/accessingLicenses.md#license-list-table-of-contents
class LicenseMetadataModel(BaseModel):
    licenseId: str = Field(title="Identifier", description="SPDX Identifier", example="Apache-2.0")
    name: str = Field(title="Name", description="Full name of the license", example="Apache License 2.0")
    reference: str = Field(
        title="Reference", description="Reference to the HTML format for the license file", example="./Apache-2.0.html"
    )
    referenceNumber: int = Field(
        title="Reference number", description="*Deprecated* - this field is generated and is no longer in use"
    )
    isDeprecatedLicenseId: bool = Field(
        title="Deprecated License", description="True if the entire license is deprecated", example=False
    )
    isOsiApproved: bool = Field(
        title="OSI approved",
        description="Indicates if the [OSI](https://opensource.org/) has approved the license",
        example=True,
    )
    seeAlso: List[HttpUrl] = Field(
        title="Reference URLs", description="Cross reference URL pointing to additional copies of the license"
    )
    detailsUrl: HttpUrl = Field(
        title="Details URL",
        description="URL to the SPDX json details for this license",
        example="http://spdx.org/licenses/Apache-2.0.json",
    )
    recommended: bool = Field(title="Recommended", description="True if this license is recommended to be used")
    url: HttpUrl = Field(title="URL", description="License URL", example="http://www.apache.org/licenses/LICENSE-2.0")
    spdxUrl: HttpUrl = Field(title="SPDX URL", example="https://spdx.org/licenses/Apache-2.0.html")


# https://docs.google.com/document/d/16vnRtDjrx5eHSl4jXs2vMaDTI6luyyLzU6xMvRHsnbI/edit#heading=h.1pihjj16olz2
RECOMMENDED_LICENSES = [
    "Apache-2.0",
    "Artistic-2.0",
    "BSD-2-Clause",
    "BSD-3-Clause",
    "CC-BY-4.0",
    "CC-BY-SA-4.0",
    "CC0-1.0",
    "EPL-2.0",
    "AGPL-3.0-or-later",
    "GPL-3.0-or-later",
    "MIT",
    "MPL-2.0",
    "PDDL-1.0",
]
SPDX_LICENSES_STRING = resource_string(__package__, "licenses.json")
SPDX_LICENSES = json.loads(SPDX_LICENSES_STRING)
for license in SPDX_LICENSES["licenses"]:
    license["recommended"] = license["licenseId"] in RECOMMENDED_LICENSES
    license["spdxUrl"] = f"https://spdx.org/licenses/{license['reference'][len('./'):]}"
    seeAlso = license.get("seeAlso", [])
    if len(seeAlso) > 0:
        url = seeAlso[0]
    else:
        url = license["spdxUrl"]
    license["url"] = url


class LicensesManager:
    def __init__(self):
        by_index = {}
        for spdx_license in self.index():
            by_index[spdx_license["licenseId"]] = spdx_license
            by_index[spdx_license["detailsUrl"]] = spdx_license
            for seeAlso in spdx_license.get("seeAlso", []):
                by_index[seeAlso] = spdx_license
        self._by_index = by_index

    def index(self):
        return SPDX_LICENSES["licenses"]

    def get(self, uri):
        if uri in self._by_index:
            return self._by_index[uri]
        else:
            log.warning(f"Unknown license URI encountered [{uri}]")
        return {"url": uri}

    def get_licenses(self) -> List[LicenseMetadataModel]:
        return SPDX_LICENSES["licenses"]

    def get_license_by_id(self, id: str) -> LicenseMetadataModel:
        license = self.get(id)
        if license.get("licenseId", None) is None:
            raise exceptions.ObjectNotFound(f"License '{id}' not found")
        return LicenseMetadataModel(**license)



import json
import logging
import os
import re
from abc import abstractmethod
from json import dumps
from typing import (
    Any,
    cast,
    Dict,
    List,
    Set,
    TYPE_CHECKING,
    Union,
)

from galaxy import model
from galaxy.exceptions import ItemAccessibilityException
from galaxy.job_execution.actions.post import ActionBox
from galaxy.model import (
    LibraryDatasetDatasetAssociation,
    WorkflowRequestInputParameter,
)
from galaxy.model.dataset_collections.builder import CollectionBuilder
from galaxy.model.none_like import NoneDataset
from galaxy.objectstore import ObjectStorePopulator
from galaxy.tools.parameters import update_dataset_ids
from galaxy.tools.parameters.basic import (
    DataCollectionToolParameter,
    DataToolParameter,
    RuntimeValue,
)
from galaxy.tools.parameters.wrapped import WrappedParameters
from galaxy.util import ExecutionTimer
from galaxy.util.template import fill_template

if TYPE_CHECKING:
    from galaxy.model import DatasetInstance

log = logging.getLogger(__name__)


class ToolExecutionCache:
    """An object mean to cache calculation caused by repeatedly evaluting
    the same tool by the same user with slightly different parameters.
    """

    def __init__(self, trans):
        self.trans = trans
        self.current_user_roles = trans.get_current_user_roles()
        self.chrom_info = {}
        self.cached_collection_elements = {}

    def get_chrom_info(self, tool_id, input_dbkey):
        genome_builds = self.trans.app.genome_builds
        custom_build_hack_get_len_from_fasta_conversion = tool_id != "CONVERTER_fasta_to_len"
        if custom_build_hack_get_len_from_fasta_conversion and input_dbkey in self.chrom_info:
            return self.chrom_info[input_dbkey]

        chrom_info_pair = genome_builds.get_chrom_info(
            input_dbkey,
            trans=self.trans,
            custom_build_hack_get_len_from_fasta_conversion=custom_build_hack_get_len_from_fasta_conversion,
        )
        if custom_build_hack_get_len_from_fasta_conversion:
            self.chrom_info[input_dbkey] = chrom_info_pair

        return chrom_info_pair


class ToolAction:
    """
    The actions to be taken when a tool is run (after parameters have
    been converted and validated).
    """

    @abstractmethod
    def execute(self, tool, trans, incoming=None, set_output_hid=True, **kwargs):
        pass


class DefaultToolAction(ToolAction):
    """Default tool action is to run an external command"""

    produces_real_jobs = True

    def _collect_input_datasets(
        self,
        tool,
        param_values,
        trans,
        history,
        current_user_roles=None,
        dataset_collection_elements=None,
        collection_info=None,
    ):
        """
        Collect any dataset inputs from incoming. Returns a mapping from
        parameter name to Dataset instance for each tool parameter that is
        of the DataToolParameter type.
        """
        if current_user_roles is None:
            current_user_roles = trans.get_current_user_roles()
        input_datasets = {}
        all_permissions: Dict[str, Set[str]] = {}

        def record_permission(action, role_id):
            if action not in all_permissions:
                all_permissions[action] = set()
            all_permissions[action].add(role_id)

        def visitor(input, value, prefix, parent=None, **kwargs):
            def process_dataset(data, formats=None):
                if not data or isinstance(data, RuntimeValue):
                    return None
                if formats is None:
                    formats = input.formats

                data = getattr(data, "hda", data)

                direct_match, target_ext, converted_dataset = data.find_conversion_destination(formats)
                if not direct_match and target_ext:
                    if converted_dataset:
                        data = converted_dataset
                    else:
                        data = data.get_converted_dataset(trans, target_ext, target_context=parent, history=history)

                input_name = prefix + input.name
                # Checked security of whole collection all at once if mapping over this input, else
                # fetch dataset details for this input from the database.
                if collection_info and collection_info.is_mapped_over(input_name):
                    action_tuples = collection_info.map_over_action_tuples(input_name)
                    if not trans.user_is_admin and not trans.app.security_agent.can_access_datasets(
                        current_user_roles, action_tuples
                    ):
                        raise ItemAccessibilityException(
                            "User does not have permission to use a dataset provided for input."
                        )
                    for action, role_id in action_tuples:
                        record_permission(action, role_id)
                else:
                    if not trans.user_is_admin and not trans.app.security_agent.can_access_dataset(
                        current_user_roles, data.dataset
                    ):
                        raise ItemAccessibilityException(
                            f"User does not have permission to use dataset ({data.name}) provided for input."
                        )
                    permissions = trans.app.security_agent.get_permissions(data.dataset)
                    for action, roles in permissions.items():
                        for role in roles:
                            record_permission(action.action, model.cached_id(role))
                return data

            if isinstance(input, DataToolParameter):
                if isinstance(value, list):
                    # If there are multiple inputs with the same name, they
                    # are stored as name1, name2, ...
                    for i, v in enumerate(value):
                        processed_dataset = process_dataset(v)
                        if i == 0:
                            # Allow copying metadata to output, first item will be source.
                            input_datasets[prefix + input.name] = processed_dataset
                        input_datasets[prefix + input.name + str(i + 1)] = processed_dataset
                        conversions = []
                        for conversion_name, conversion_extensions, conversion_datatypes in input.conversions:
                            new_data = process_dataset(
                                input_datasets[prefix + input.name + str(i + 1)], conversion_datatypes
                            )
                            if not new_data or new_data.datatype.matches_any(conversion_datatypes):
                                input_datasets[prefix + conversion_name + str(i + 1)] = new_data
                                conversions.append((conversion_name, new_data))
                            else:
                                raise Exception(
                                    f"A path for explicit datatype conversion has not been found: {input_datasets[prefix + input.name + str(i + 1)].extension} --/--> {conversion_extensions}"
                                )
                        if parent:
                            parent[input.name][i] = input_datasets[prefix + input.name + str(i + 1)]
                            for conversion_name, conversion_data in conversions:
                                # allow explicit conversion to be stored in job_parameter table
                                parent[conversion_name][
                                    i
                                ] = conversion_data.id  # a more robust way to determine JSONable value is desired
                        else:
                            param_values[input.name][i] = input_datasets[prefix + input.name + str(i + 1)]
                            for conversion_name, conversion_data in conversions:
                                # allow explicit conversion to be stored in job_parameter table
                                param_values[conversion_name][
                                    i
                                ] = conversion_data.id  # a more robust way to determine JSONable value is desired
                else:
                    input_datasets[prefix + input.name] = process_dataset(value)
                    conversions = []
                    for conversion_name, conversion_extensions, conversion_datatypes in input.conversions:
                        new_data = process_dataset(input_datasets[prefix + input.name], conversion_datatypes)
                        if not new_data or new_data.datatype.matches_any(conversion_datatypes):
                            input_datasets[prefix + conversion_name] = new_data
                            conversions.append((conversion_name, new_data))
                        else:
                            raise Exception(
                                f"A path for explicit datatype conversion has not been found: {input_datasets[prefix + input.name].extension} --/--> {conversion_extensions}"
                            )
                    target_dict = parent
                    if not target_dict:
                        target_dict = param_values
                    target_dict[input.name] = input_datasets[prefix + input.name]
                    for conversion_name, conversion_data in conversions:
                        # allow explicit conversion to be stored in job_parameter table
                        target_dict[
                            conversion_name
                        ] = conversion_data.id  # a more robust way to determine JSONable value is desired
            elif isinstance(input, DataCollectionToolParameter):
                if not value:
                    return

                collection = None
                child_collection = False
                if hasattr(value, "child_collection"):
                    # if we are mapping a collection over a tool, we only require the child_collection
                    child_collection = True
                    collection = value.child_collection
                else:
                    # else the tool takes a collection as input so we need everything
                    collection = value.collection

                action_tuples = collection.dataset_action_tuples
                if not trans.user_is_admin and not trans.app.security_agent.can_access_datasets(
                    current_user_roles, action_tuples
                ):
                    raise ItemAccessibilityException(
                        "User does not have permission to use a dataset provided for input."
                    )
                for action, role_id in action_tuples:
                    record_permission(action, role_id)

                _, extensions = collection.dataset_states_and_extensions_summary
                conversion_required = False
                for ext in extensions:
                    if ext:
                        datatype = trans.app.datatypes_registry.get_datatype_by_extension(ext)
                        if not datatype.matches_any(input.formats):
                            conversion_required = True
                            break
                processed_dataset_dict = {}
                for i, v in enumerate(collection.dataset_instances):
                    processed_dataset = None
                    if conversion_required:
                        processed_dataset = process_dataset(v)
                        if processed_dataset is not v:
                            processed_dataset_dict[v] = processed_dataset
                    input_datasets[prefix + input.name + str(i + 1)] = processed_dataset or v
                if conversion_required:
                    collection_type_description = (
                        trans.app.dataset_collection_manager.collection_type_descriptions.for_collection_type(
                            collection.collection_type
                        )
                    )
                    collection_builder = CollectionBuilder(collection_type_description)
                    collection_builder.replace_elements_in_collection(
                        template_collection=collection,
                        replacement_dict=processed_dataset_dict,
                    )
                    new_collection = collection_builder.build()
                    if child_collection:
                        value.child_collection = new_collection
                    else:
                        value.collection = new_collection

        tool.visit_inputs(param_values, visitor)
        return input_datasets, all_permissions

    def collect_input_dataset_collections(self, tool, param_values):
        def append_to_key(the_dict, key, value):
            if key not in the_dict:
                the_dict[key] = []
            the_dict[key].append(value)

        input_dataset_collections: Dict[str, str] = {}

        def visitor(input, value, prefix, parent=None, prefixed_name=None, **kwargs):
            if isinstance(input, DataToolParameter):
                values = value
                if not isinstance(values, list):
                    values = [value]
                for i, value in enumerate(values):
                    if isinstance(value, model.HistoryDatasetCollectionAssociation) or isinstance(
                        value, model.DatasetCollectionElement
                    ):
                        append_to_key(input_dataset_collections, prefixed_name, (value, True))
                        target_dict = parent
                        if not target_dict:
                            target_dict = param_values
                        # This is just a DataToolParameter, so replace this
                        # collection with individual datasets. Database will still
                        # record collection which should be enought for workflow
                        # extraction and tool rerun.
                        if isinstance(value, model.DatasetCollectionElement):
                            if value.child_collection:
                                # if we are mapping a collection over a tool, we only require the child_collection
                                dataset_instances = value.child_collection.dataset_instances
                            else:
                                continue
                        else:
                            # else the tool takes a collection as input so we need everything
                            dataset_instances = value.collection.dataset_instances
                        if i == 0:
                            target_dict[input.name] = []
                        target_dict[input.name].extend(dataset_instances)
            elif isinstance(input, DataCollectionToolParameter):
                append_to_key(input_dataset_collections, prefix + input.name, (value, False))

        tool.visit_inputs(param_values, visitor)
        return input_dataset_collections

    def _check_access(self, tool, trans):
        assert tool.allow_user_access(trans.user), f"User ({trans.user}) is not allowed to access this tool."

    def _collect_inputs(self, tool, trans, incoming, history, current_user_roles, collection_info):
        """Collect history as well as input datasets and collections."""
        # Set history.
        if not history:
            history = tool.get_default_history_by_trans(trans, create=True)

        # Track input dataset collections - but replace with simply lists so collect
        # input datasets can process these normally.
        inp_dataset_collections = self.collect_input_dataset_collections(tool, incoming)
        # Collect any input datasets from the incoming parameters
        inp_data, all_permissions = self._collect_input_datasets(
            tool,
            incoming,
            trans,
            history=history,
            current_user_roles=current_user_roles,
            collection_info=collection_info,
        )

        preserved_tags = {}
        preserved_hdca_tags = {}
        # grab tags from incoming HDAs
        for data in inp_data.values():
            if not data:
                continue
            for tag in data.auto_propagated_tags:
                preserved_tags[tag.value] = tag
        # grab tags from incoming HDCAs
        for collection_pairs in inp_dataset_collections.values():
            for collection, _ in collection_pairs:
                # if sub-collection mapping, this will be an DC not an HDCA
                # (e.g. part of collection not a collection instance) and thus won't have tags.
                if hasattr(collection, "tags"):
                    for tag in collection.auto_propagated_tags:
                        preserved_hdca_tags[tag.value] = tag
        preserved_tags.update(preserved_hdca_tags)
        return history, inp_data, inp_dataset_collections, preserved_tags, preserved_hdca_tags, all_permissions

    def execute(
        self,
        tool,
        trans,
        incoming=None,
        return_job=False,
        set_output_hid=True,
        history=None,
        job_params=None,
        rerun_remap_job_id=None,
        execution_cache=None,
        dataset_collection_elements=None,
        completed_job=None,
        collection_info=None,
        job_callback=None,
        flush_job=True,
        skip=False,
    ):
        """
        Executes a tool, creating job and tool outputs, associating them, and
        submitting the job to the job queue. If history is not specified, use
        trans.history as destination for tool's output datasets.
        """
        trans.check_user_activation()
        incoming = incoming or {}
        self._check_access(tool, trans)
        app = trans.app
        if execution_cache is None:
            execution_cache = ToolExecutionCache(trans)
        current_user_roles = execution_cache.current_user_roles
        (
            history,
            inp_data,
            inp_dataset_collections,
            preserved_tags,
            preserved_hdca_tags,
            all_permissions,
        ) = self._collect_inputs(tool, trans, incoming, history, current_user_roles, collection_info)
        # Build name for output datasets based on tool name and input names
        on_text = self._get_on_text(inp_data)

        # format='input" previously would give you a random extension from
        # the input extensions, now it should just give "input" as the output
        # format.
        input_ext = "data" if tool.profile < 16.04 else "input"
        input_dbkey = incoming.get("dbkey", "?")
        for name, data in reversed(list(inp_data.items())):
            if not data:
                data = NoneDataset(datatypes_registry=app.datatypes_registry)
                continue

            # Convert LDDA to an HDA.
            if isinstance(data, LibraryDatasetDatasetAssociation) and not completed_job:
                data = data.to_history_dataset_association(None)
                inp_data[name] = data

            if tool.profile < 16.04:
                input_ext = data.ext

            if data.dbkey not in [None, "?"]:
                input_dbkey = data.dbkey

            identifier = getattr(data, "element_identifier", None)
            if identifier is not None:
                incoming[f"{name}|__identifier__"] = identifier

        # Collect chromInfo dataset and add as parameters to incoming
        (chrom_info, db_dataset) = execution_cache.get_chrom_info(tool.id, input_dbkey)

        if db_dataset:
            inp_data.update({"chromInfo": db_dataset})
        incoming["chromInfo"] = chrom_info

        if not completed_job:
            # Determine output dataset permission/roles list
            if all_permissions:
                output_permissions = app.security_agent.guess_derived_permissions(all_permissions)
            else:
                # No valid inputs, we will use history defaults
                output_permissions = app.security_agent.history_get_default_permissions(history)

        # Add the dbkey to the incoming parameters
        incoming["dbkey"] = input_dbkey
        incoming["__input_ext"] = input_ext
        # wrapped params are used by change_format action and by output.label; only perform this wrapping once, as needed
        wrapped_params = self._wrapped_params(trans, tool, incoming, inp_data)

        out_data: Dict[str, "DatasetInstance"] = {}
        input_collections = {k: v[0][0] for k, v in inp_dataset_collections.items()}
        output_collections = OutputCollections(
            trans,
            history,
            tool=tool,
            tool_action=self,
            input_collections=input_collections,
            dataset_collection_elements=dataset_collection_elements,
            on_text=on_text,
            incoming=incoming,
            params=wrapped_params.params,
            job_params=job_params,
            tags=preserved_tags,
            hdca_tags=preserved_hdca_tags,
        )

        async_tool = tool.tool_type == "data_source_async"

        def handle_output(name, output, hidden=None):
            if async_tool and name in incoming:
                # HACK: output data has already been created as a result of the async controller
                dataid = incoming[name]
                data = trans.sa_session.query(app.model.HistoryDatasetAssociation).get(dataid)
                assert data is not None
                out_data[name] = data
            else:
                ext = determine_output_format(
                    output,
                    wrapped_params.params,
                    inp_data,
                    inp_dataset_collections,
                    input_ext,
                    python_template_version=tool.python_template_version,
                    execution_cache=execution_cache,
                )
                create_datasets = True
                dataset = None

                if completed_job:
                    for output_dataset in completed_job.output_datasets:
                        if output_dataset.name == name:
                            create_datasets = False
                            completed_data = output_dataset.dataset
                            dataset = output_dataset.dataset.dataset
                            break

                data = app.model.HistoryDatasetAssociation(
                    extension=ext, dataset=dataset, create_dataset=create_datasets, flush=False
                )
                if create_datasets:
                    from_work_dir = output.from_work_dir
                    if from_work_dir is not None:
                        data.dataset.created_from_basename = os.path.basename(from_work_dir)
                if hidden is None:
                    hidden = output.hidden
                if not hidden and dataset_collection_elements is not None:  # Mapping over a collection - hide datasets
                    hidden = True
                if hidden:
                    data.visible = False
                if dataset_collection_elements is not None and name in dataset_collection_elements:
                    dataset_collection_elements[name].hda = data
                trans.sa_session.add(data)
                if not completed_job:
                    trans.app.security_agent.set_all_dataset_permissions(
                        data.dataset, output_permissions, new=True, flush=False
                    )
            data.copy_tags_to(preserved_tags.values())

            # This may not be necessary with the new parent/child associations
            data.designation = name
            # Copy metadata from one of the inputs if requested.

            # metadata source can be either a string referencing an input
            # or an actual object to copy.
            metadata_source = output.metadata_source
            if metadata_source:
                if isinstance(metadata_source, str):
                    metadata_source = inp_data.get(metadata_source)

            if metadata_source is not None:
                data.init_meta(copy_from=metadata_source)
            else:
                data.init_meta()
            # Take dbkey from LAST input
            data.dbkey = str(input_dbkey)
            # Set state
            if completed_job:
                data.blurb = completed_data.blurb
                data.peek = completed_data.peek
                data._metadata = completed_data._metadata
            else:
                data.blurb = "queued"
            # Set output label
            data.name = self.get_output_name(
                output, data, tool, on_text, trans, incoming, history, wrapped_params.params, job_params
            )
            # Store output
            out_data[name] = data
            if output.actions:
                # Apply pre-job tool-output-dataset actions; e.g. setting metadata, changing format
                output_action_params = dict(out_data)
                output_action_params.update(wrapped_params.params)
                output_action_params["__python_template_version__"] = tool.python_template_version
                output.actions.apply_action(data, output_action_params)
            # Flush all datasets at once.
            return data

        child_dataset_names = set()

        for name, output in tool.outputs.items():
            if not filter_output(tool, output, incoming):
                handle_output_timer = ExecutionTimer()
                if output.collection:
                    if completed_job and dataset_collection_elements and name in dataset_collection_elements:
                        # Output collection is mapped over and has already been copied from original job
                        continue
                    collections_manager = app.dataset_collection_manager
                    element_identifiers: List[Dict[str, Union[str, List[Dict[str, Union[str, List[Any]]]]]]] = []
                    # mypy doesn't yet support recursive type definitions
                    known_outputs = output.known_outputs(input_collections, collections_manager.type_registry)
                    # Just to echo TODO elsewhere - this should be restructured to allow
                    # nested collections.
                    for output_part_def in known_outputs:
                        # Add elements to top-level collection, unless nested...
                        current_element_identifiers = element_identifiers
                        current_collection_type = output.structure.collection_type

                        for parent_id in output_part_def.parent_ids or []:
                            # TODO: replace following line with formal abstractions for doing this.
                            current_collection_type = ":".join(current_collection_type.split(":")[1:])
                            name_to_index = {
                                value["name"]: index for (index, value) in enumerate(current_element_identifiers)
                            }
                            if parent_id not in name_to_index:
                                if parent_id not in current_element_identifiers:
                                    index = len(current_element_identifiers)
                                    current_element_identifiers.append(
                                        dict(
                                            name=parent_id,
                                            collection_type=current_collection_type,
                                            src="new_collection",
                                            element_identifiers=[],
                                        )
                                    )
                                else:
                                    index = name_to_index[parent_id]
                            current_element_identifiers = cast(
                                List[
                                    Dict[
                                        str,
                                        Union[str, List[Dict[str, Union[str, List[Any]]]]],
                                    ]
                                ],
                                current_element_identifiers[index]["element_identifiers"],
                            )

                        effective_output_name = output_part_def.effective_output_name
                        child_dataset_names.add(effective_output_name)
                        element = handle_output(effective_output_name, output_part_def.output_def, hidden=True)
                        history.stage_addition(element)
                        # TODO: this shouldn't exist in the top-level of the history at all
                        # but for now we are still working around that by hiding the contents
                        # there.
                        # Following hack causes dataset to no be added to history...
                        trans.sa_session.add(element)
                        current_element_identifiers.append(
                            {
                                "__object__": element,
                                "name": output_part_def.element_identifier,
                            }
                        )

                    if output.dynamic_structure:
                        assert not element_identifiers  # known_outputs must have been empty
                        element_kwds = dict(elements=collections_manager.ELEMENTS_UNINITIALIZED)
                    else:
                        element_kwds = dict(element_identifiers=element_identifiers)
                    output_collections.create_collection(
                        output=output, name=name, completed_job=completed_job, **element_kwds
                    )
                    log.info(f"Handled collection output named {name} for tool {tool.id} {handle_output_timer}")
                else:
                    handle_output(name, output)
                    log.info(f"Handled output named {name} for tool {tool.id} {handle_output_timer}")

        add_datasets_timer = tool.app.execution_timer_factory.get_timer(
            "internals.galaxy.tools.actions.add_datasets",
            "Added output datasets to history",
        )
        # Add all the top-level (non-child) datasets to the history unless otherwise specified
        for name, data in out_data.items():
            if name not in incoming and name not in child_dataset_names:
                # don't add already existing datasets, i.e. async created
                history.stage_addition(data)
        history.add_pending_items(set_output_hid=set_output_hid)

        log.info(add_datasets_timer)
        job_setup_timer = ExecutionTimer()
        # Create the job object
        job, galaxy_session = self._new_job_for_session(trans, tool, history)
        if skip:
            job.state = job.states.SKIPPED
            for output_collection in output_collections.out_collections.values():
                output_collection.mark_as_populated()
            object_store_populator = ObjectStorePopulator(trans.app, trans.user)
            for data in out_data.values():
                object_store_populator.set_object_store_id(data)
                data.extension = "expression.json"
                data.state = "ok"
                with open(data.dataset.file_name, "w") as out:
                    out.write(json.dumps(None))
        self._record_inputs(trans, tool, job, incoming, inp_data, inp_dataset_collections)
        self._record_outputs(job, out_data, output_collections)
        # execute immediate post job actions and associate post job actions that are to be executed after the job is complete
        if job_callback:
            job_callback(job)
        if job_params:
            job.params = dumps(job_params)
        if completed_job:
            job.set_copied_from_job_id(completed_job.id)
        trans.sa_session.add(job)
        # Remap any outputs if this is a rerun and the user chose to continue dependent jobs
        # This functionality requires tracking jobs in the database.
        if app.config.track_jobs_in_database and rerun_remap_job_id is not None:
            # Need to flush here so that referencing outputs by id works
            session = trans.sa_session()
            try:
                session.expire_on_commit = False
                session.flush()
            finally:
                session.expire_on_commit = True
            self._remap_job_on_rerun(
                trans=trans,
                galaxy_session=galaxy_session,
                rerun_remap_job_id=rerun_remap_job_id,
                current_job=job,
                out_data=out_data,
            )
        log.info(f"Setup for job {job.log_str()} complete, ready to be enqueued {job_setup_timer}")

        # Some tools are not really executable, but jobs are still created for them ( for record keeping ).
        # Examples include tools that redirect to other applications ( epigraph ).  These special tools must
        # include something that can be retrieved from the params ( e.g., REDIRECT_URL ) to keep the job
        # from being queued.
        if "REDIRECT_URL" in incoming:
            # Get the dataset - there should only be 1
            for name in inp_data.keys():
                dataset = inp_data[name]
            redirect_url = tool.parse_redirect_url(dataset, incoming)
            # GALAXY_URL should be include in the tool params to enable the external application
            # to send back to the current Galaxy instance
            GALAXY_URL = incoming.get("GALAXY_URL", None)
            assert GALAXY_URL is not None, "GALAXY_URL parameter missing in tool config."
            redirect_url += f"&GALAXY_URL={GALAXY_URL}"
            # Job should not be queued, so set state to ok
            job.set_state(app.model.Job.states.OK)
            job.info = f"Redirected to: {redirect_url}"
            trans.sa_session.add(job)
            trans.sa_session.flush()
            trans.response.send_redirect(redirect_url)
        else:
            if flush_job:
                # Set HID and add to history.
                job_flush_timer = ExecutionTimer()
                trans.sa_session.flush()
                log.info(f"Flushed transaction for job {job.log_str()} {job_flush_timer}")

        return job, out_data, history

    def _remap_job_on_rerun(self, trans, galaxy_session, rerun_remap_job_id, current_job, out_data):
        """
        Re-connect dependent datasets for a job that is being rerun (because it failed initially).

        If a job fails, the user has the option to try the job again with changed parameters.
        To be able to resume jobs that depend on this jobs output datasets we change the dependent's job
        input datasets to be those of the job that is being rerun.
        """
        try:
            old_job = trans.sa_session.query(trans.app.model.Job).get(rerun_remap_job_id)
            assert old_job is not None, f"({rerun_remap_job_id}/{current_job.id}): Old job id is invalid"
            assert (
                old_job.tool_id == current_job.tool_id
            ), f"({old_job.id}/{current_job.id}): Old tool id ({old_job.tool_id}) does not match rerun tool id ({current_job.tool_id})"
            if trans.user is not None:
                assert (
                    old_job.user_id == trans.user.id
                ), f"({old_job.id}/{current_job.id}): Old user id ({old_job.user_id}) does not match rerun user id ({trans.user.id})"
            elif trans.user is None and type(galaxy_session) == trans.model.GalaxySession:
                assert (
                    old_job.session_id == galaxy_session.id
                ), f"({old_job.id}/{current_job.id}): Old session id ({old_job.session_id}) does not match rerun session id ({galaxy_session.id})"
            else:
                raise Exception(f"({old_job.id}/{current_job.id}): Remapping via the API is not (yet) supported")
            # Start by hiding current job outputs before taking over the old job's (implicit) outputs.
            current_job.hide_outputs(flush=False)
            # Duplicate PJAs before remap.
            for pjaa in old_job.post_job_actions:
                current_job.add_post_job_action(pjaa.post_job_action)
            if old_job.workflow_invocation_step:
                replacement_dict = {}
                for parameter in old_job.workflow_invocation_step.workflow_invocation.input_parameters:
                    if parameter.type == WorkflowRequestInputParameter.types.REPLACEMENT_PARAMETERS:
                        replacement_dict[parameter.name] = parameter.value
                for pja in old_job.workflow_invocation_step.workflow_step.post_job_actions:
                    # execute immediate actions here, with workflow context.
                    if pja.action_type in ActionBox.immediate_actions:
                        ActionBox.execute(trans.app, trans.sa_session, pja, current_job, replacement_dict)
            for p in old_job.parameters:
                if p.name.endswith("|__identifier__"):
                    current_job.parameters.append(p.copy())
            remapped_hdas = self.__remap_data_inputs(old_job=old_job, current_job=current_job)
            for jtod in old_job.output_datasets:
                for job_to_remap, jtid in [(jtid.job, jtid) for jtid in jtod.dataset.dependent_jobs]:
                    if (trans.user is not None and job_to_remap.user_id == trans.user.id) or (
                        trans.user is None and job_to_remap.session_id == galaxy_session.id
                    ):
                        self.__remap_parameters(job_to_remap, jtid, jtod, out_data)
                        trans.sa_session.add(job_to_remap)
                        trans.sa_session.add(jtid)
                        job_to_remap.resume()
                jtod.dataset.visible = False
                trans.sa_session.add(jtod)
            for jtodc in old_job.output_dataset_collection_instances:
                # Update JobToOutputDatasetCollectionAssociation to the current job
                jtodc.job = current_job
                hdca = jtodc.dataset_collection_instance
                hdca.collection.replace_failed_elements(remapped_hdas)
                if hdca.implicit_collection_jobs:
                    for job in hdca.implicit_collection_jobs.jobs:
                        if job.job_id == old_job.id:
                            job.job_id = current_job.id
                hdca.update()
            for jtoidca in old_job.output_dataset_collections:
                jtoidca.dataset_collection.replace_failed_elements(remapped_hdas)
        except Exception:
            log.exception("Cannot remap rerun dependencies.")

    def __remap_data_inputs(self, old_job, current_job):
        """Record output datasets from old_job and build a dictionary that maps the old output HDAs to the new output HDAs."""
        remapped_hdas = {}
        old_output_datasets = {jtod.name: jtod.dataset for jtod in old_job.output_datasets}
        for jtod in current_job.output_datasets:
            remapped_hdas[old_output_datasets[jtod.name]] = jtod.dataset
        return remapped_hdas

    def __remap_parameters(self, job_to_remap, jtid, jtod, out_data):
        input_values = {p.name: json.loads(p.value) for p in job_to_remap.parameters if p.value is not None}
        old_dataset_id = jtod.dataset_id
        new_dataset_id = out_data[jtod.name].id
        input_values = update_dataset_ids(input_values, {old_dataset_id: new_dataset_id}, src="hda")
        for p in job_to_remap.parameters:
            if p.name in input_values:
                p.value = json.dumps(input_values[p.name])
        jtid.dataset = out_data[jtod.name]
        jtid.dataset.hid = jtod.dataset.hid
        log.info(f"Job {job_to_remap.id} input HDA {jtod.dataset.id} remapped to new HDA {jtid.dataset.id}")

    def _wrapped_params(self, trans, tool, incoming, input_datasets=None):
        wrapped_params = WrappedParameters(trans, tool, incoming, input_datasets=input_datasets)
        return wrapped_params

    def _get_on_text(self, inp_data):
        input_names = []
        for data in reversed(list(inp_data.values())):
            if getattr(data, "hid", None):
                input_names.append(f"data {data.hid}")

        return on_text_for_names(input_names)

    def _new_job_for_session(self, trans, tool, history):
        job = trans.app.model.Job()
        job.galaxy_version = trans.app.config.version_major
        galaxy_session = None

        if hasattr(trans, "get_galaxy_session"):
            galaxy_session = trans.get_galaxy_session()
            # If we're submitting from the API, there won't be a session.
            if type(galaxy_session) == trans.model.GalaxySession:
                job.session_id = model.cached_id(galaxy_session)
        if trans.user is not None:
            job.user_id = model.cached_id(trans.user)
        if history:
            job.history_id = model.cached_id(history)
        job.tool_id = tool.id
        try:
            # For backward compatibility, some tools may not have versions yet.
            job.tool_version = tool.version
        except AttributeError:
            job.tool_version = "1.0.0"
        job.dynamic_tool = tool.dynamic_tool
        return job, galaxy_session

    def _record_inputs(self, trans, tool, job, incoming, inp_data, inp_dataset_collections):
        # FIXME: Don't need all of incoming here, just the defined parameters
        #        from the tool. We need to deal with tools that pass all post
        #        parameters to the command as a special case.
        reductions: Dict[str, List[str]] = {}
        for name, dataset_collection_info_pairs in inp_dataset_collections.items():
            for dataset_collection, reduced in dataset_collection_info_pairs:
                if reduced:
                    if name not in reductions:
                        reductions[name] = []
                    reductions[name].append(dataset_collection)

                # TODO: verify can have multiple with same name, don't want to lose traceability
                if isinstance(dataset_collection, model.HistoryDatasetCollectionAssociation):
                    job.add_input_dataset_collection(name, dataset_collection)
                elif isinstance(dataset_collection, model.DatasetCollectionElement):
                    job.add_input_dataset_collection_element(name, dataset_collection)

        # If this an input collection is a reduction, we expanded it for dataset security, type
        # checking, and such, but the persisted input must be the original collection
        # so we can recover things like element identifier during tool command evaluation.
        def restore_reduction_visitor(input, value, prefix, parent=None, prefixed_name=None, **kwargs):
            if prefixed_name in reductions and isinstance(input, DataToolParameter):
                target_dict = parent
                if not target_dict:
                    target_dict = incoming

                target_dict[input.name] = []
                for reduced_collection in reductions[prefixed_name]:
                    if hasattr(reduced_collection, "child_collection"):
                        target_dict[input.name].append({"id": model.cached_id(reduced_collection), "src": "dce"})
                    else:
                        target_dict[input.name].append({"id": model.cached_id(reduced_collection), "src": "hdca"})

        if reductions:
            tool.visit_inputs(incoming, restore_reduction_visitor)

        for name, value in tool.params_to_strings(incoming, trans.app).items():
            job.add_parameter(name, value)
        self._record_input_datasets(trans, job, inp_data)

    def _record_outputs(self, job, out_data, output_collections):
        out_collections = output_collections.out_collections
        out_collection_instances = output_collections.out_collection_instances
        for name, dataset in out_data.items():
            job.add_output_dataset(name, dataset)
        for name, dataset_collection in out_collections.items():
            job.add_implicit_output_dataset_collection(name, dataset_collection)
        for name, dataset_collection_instance in out_collection_instances.items():
            job.add_output_dataset_collection(name, dataset_collection_instance)
            dataset_collection_instance.job = job

    def _record_input_datasets(self, trans, job, inp_data):
        for name, dataset in inp_data.items():
            # TODO: figure out why can't pass dataset_id here.
            job.add_input_dataset(name, dataset=dataset)

    def get_output_name(
        self,
        output,
        dataset=None,
        tool=None,
        on_text=None,
        trans=None,
        incoming=None,
        history=None,
        params=None,
        job_params=None,
    ):
        if output.label:
            params["tool"] = tool
            params["on_string"] = on_text
            return fill_template(output.label, context=params, python_template_version=tool.python_template_version)
        else:
            return self._get_default_data_name(
                dataset,
                tool,
                on_text=on_text,
                trans=trans,
                incoming=incoming,
                history=history,
                params=params,
                job_params=job_params,
            )

    def _get_default_data_name(
        self, dataset, tool, on_text=None, trans=None, incoming=None, history=None, params=None, job_params=None, **kwd
    ):
        name = tool.name
        if on_text:
            name += f" on {on_text}"
        return name


class OutputCollections:
    """Keeps track of collections (DC or HDCA) created by actions.

    Actions do fairly different things depending on whether we are creating
    just part of an collection or a whole output collection (mapping_over_collection
    parameter).
    """

    def __init__(
        self,
        trans,
        history,
        tool,
        tool_action,
        input_collections,
        dataset_collection_elements,
        on_text,
        incoming,
        params,
        job_params,
        tags,
        hdca_tags,
    ):
        self.trans = trans
        self.tag_handler = trans.app.tag_handler.create_tag_handler_session()
        self.history = history
        self.tool = tool
        self.tool_action = tool_action
        self.input_collections = input_collections
        self.dataset_collection_elements = dataset_collection_elements
        self.on_text = on_text
        self.incoming = incoming
        self.params = params
        self.job_params = job_params
        self.out_collections = {}
        self.out_collection_instances = {}
        self.tags = tags  # all inherited tags
        self.hdca_tags = hdca_tags  # only tags inherited from input HDCAs

    def create_collection(
        self, output, name, collection_type=None, completed_job=None, propagate_hda_tags=True, **element_kwds
    ):
        input_collections = self.input_collections
        collections_manager = self.trans.app.dataset_collection_manager
        collection_type = collection_type or output.structure.collection_type
        if collection_type is None:
            collection_type_source = output.structure.collection_type_source
            if collection_type_source is None:
                # TODO: Not a new problem, but this should be determined
                # sooner.
                raise Exception("Could not determine collection type to create.")
            if collection_type_source not in input_collections:
                raise Exception(f"Could not find collection type source with name [{collection_type_source}].")

            # Using the collection_type_source string we get the DataCollectionToolParameter
            data_param = self.tool.inputs
            groups = collection_type_source.split("|")
            for group in groups:
                values = group.split("_")
                if values[-1].isdigit():
                    key = "_".join(values[0:-1])
                    # We don't care about the repeat index, we just need to find the correct DataCollectionToolParameter
                else:
                    key = group
                if isinstance(data_param, dict):
                    data_param = data_param.get(key)
                else:
                    data_param = data_param.inputs.get(key)
            collection_type_description = data_param._history_query(self.trans).can_map_over(
                input_collections[collection_type_source]
            )
            if collection_type_description:
                collection_type = collection_type_description.collection_type
            else:
                collection_type = input_collections[collection_type_source].collection.collection_type

        if "elements" in element_kwds:

            def check_elements(elements):
                if hasattr(elements, "items"):  # else it is ELEMENTS_UNINITIALIZED object.
                    for value in elements.values():
                        # Either a HDA (if) or a DatasetCollection or a recursive dict.
                        if getattr(value, "history_content_type", None) == "dataset":
                            assert value.history is not None or value.history_id is not None
                        elif hasattr(value, "dataset_instances"):
                            for dataset in value.dataset_instances:
                                assert dataset.history is not None or dataset.history_id is not None
                        else:
                            assert value["src"] == "new_collection"
                            check_elements(value["elements"])

            elements = element_kwds["elements"]
            check_elements(elements)

        if self.dataset_collection_elements is not None:
            dc = collections_manager.create_dataset_collection(
                self.trans, collection_type=collection_type, **element_kwds
            )
            if name in self.dataset_collection_elements:
                self.dataset_collection_elements[name].child_collection = dc
                # self.trans.sa_session.add(self.dataset_collection_elements[name])
            self.out_collections[name] = dc
        else:
            hdca_name = self.tool_action.get_output_name(
                output,
                None,
                self.tool,
                self.on_text,
                self.trans,
                self.incoming,
                self.history,
                self.params,
                self.job_params,
            )
            hdca = collections_manager.create(
                self.trans,
                self.history,
                name=hdca_name,
                collection_type=collection_type,
                trusted_identifiers=True,
                tags=self.tags if propagate_hda_tags else self.hdca_tags,
                set_hid=False,
                flush=False,
                completed_job=completed_job,
                output_name=name,
                **element_kwds,
            )
            # name here is name of the output element - not name
            # of the hdca.
            self.history.stage_addition(hdca)
            self.out_collection_instances[name] = hdca


def on_text_for_names(input_names):
    # input_names may contain duplicates... this is because the first value in
    # multiple input dataset parameters will appear twice once as param_name
    # and once as param_name1.
    unique_names = []
    for name in input_names:
        if name not in unique_names:
            unique_names.append(name)
    input_names = unique_names

    # Build name for output datasets based on tool name and input names
    if len(input_names) == 1:
        on_text = input_names[0]
    elif len(input_names) == 2:
        on_text = "%s and %s" % tuple(input_names[0:2])
    elif len(input_names) == 3:
        on_text = "%s, %s, and %s" % tuple(input_names[0:3])
    elif len(input_names) > 3:
        on_text = "%s, %s, and others" % tuple(input_names[0:2])
    else:
        on_text = ""
    return on_text


def filter_output(tool, output, incoming):
    for filter in output.filters:
        try:
            if not eval(filter.text.strip(), globals(), incoming):
                return True  # do not create this dataset
        except Exception as e:
            log.debug(f"Tool {tool.id} output {output.name}: dataset output filter ({filter.text}) failed: {e}")
    return False


def get_ext_or_implicit_ext(hda):
    if hda.implicitly_converted_parent_datasets:
        # implicitly_converted_parent_datasets is a list of ImplicitlyConvertedDatasetAssociation
        # objects, and their type is the target_ext, so this should be correct even if there
        # are multiple ImplicitlyConvertedDatasetAssociation objects (meaning 2 datasets had been converted
        # to produce a dataset with the required datatype)
        return hda.implicitly_converted_parent_datasets[0].type
    return hda.ext


def determine_output_format(
    output,
    parameter_context,
    input_datasets,
    input_dataset_collections,
    random_input_ext,
    python_template_version="3",
    execution_cache=None,
):
    """Determines the output format for a dataset based on an abstract
    description of the output (galaxy.tool_util.parser.ToolOutput), the parameter
    wrappers, a map of the input datasets (name => HDA), and the last input
    extensions in the tool form.

    TODO: Don't deal with XML here - move this logic into ToolOutput.
    TODO: Make the input extension used deterministic instead of random.
    """
    # the type should match the input
    ext = output.format
    if ext == "input":
        if input_datasets and random_input_ext in {"data", "auto"}:
            # Probably dealing with an implicitly converted dataset
            try:
                first_input_dataset = next(iter(input_datasets.values()))
                random_input_ext = get_ext_or_implicit_ext(first_input_dataset)
            except Exception:
                pass
        ext = random_input_ext
    format_source = output.format_source
    if format_source is not None and format_source in input_datasets:
        try:
            input_dataset = input_datasets[output.format_source]
            ext = get_ext_or_implicit_ext(input_dataset)
        except Exception:
            pass
    elif format_source is not None:
        element_index = None
        collection_name = format_source
        if re.match(r"^[^\[\]]*\[[^\[\]]*\]$", format_source):
            collection_name, element_index = format_source[0:-1].split("[")
            # Treat as json to interpret "forward" vs 0 with type
            # Make it feel more like Python, single quote better in XML also.
            element_index = element_index.replace("'", '"')
            element_index = json.loads(element_index)

        if collection_name in input_dataset_collections:
            try:
                input_collection = input_dataset_collections[collection_name][0][0]
                input_collection_collection = input_collection.collection
                if element_index is None:
                    # just pick the first HDA
                    input_dataset = input_collection_collection.dataset_instances[0]
                else:
                    try:
                        input_element = input_collection_collection[element_index]
                    except KeyError:
                        if execution_cache:
                            dataset_elements = execution_cache.cached_collection_elements.get(
                                input_collection_collection.id
                            )
                            if dataset_elements is None:
                                dataset_elements = execution_cache.cached_collection_elements[
                                    input_collection_collection.id
                                ] = input_collection_collection.dataset_elements
                        else:
                            dataset_elements = input_collection_collection.dataset_elements
                        for element in dataset_elements:
                            if element.element_identifier == element_index:
                                input_element = element
                                break
                    input_dataset = input_element.element_object
                ext = get_ext_or_implicit_ext(input_dataset)
            except Exception as e:
                log.debug("Exception while trying to determine format_source: %s", e)

    # process change_format tags
    if output.change_format is not None:
        new_format_set = False
        for change_elem in output.change_format:
            for when_elem in change_elem.findall("when"):
                check = when_elem.get("input", None)
                if check is not None:
                    try:
                        if "$" not in check:
                            # allow a simple name or more complex specifications
                            check = "${%s}" % check
                        if fill_template(
                            check, context=parameter_context, python_template_version=python_template_version
                        ) == when_elem.get("value", None):
                            ext = when_elem.get("format", ext)
                    except Exception:
                        # bad tag input value; possibly referencing a param within a different conditional when block or other nonexistent grouping construct
                        continue
                else:
                    check = when_elem.get("input_dataset", None)
                    if check is not None:
                        check = input_datasets.get(check, None)
                        # At this point check is a HistoryDatasetAssociation object.
                        check_format = when_elem.get("format", ext)
                        check_value = when_elem.get("value", None)
                        check_attribute = when_elem.get("attribute", None)
                        if check is not None and check_value is not None and check_attribute is not None:
                            # See if the attribute to be checked belongs to the HistoryDatasetAssociation object.
                            if hasattr(check, check_attribute):
                                if str(getattr(check, check_attribute)) == str(check_value):
                                    ext = check_format
                                    new_format_set = True
                                    break
                            # See if the attribute to be checked belongs to the metadata associated with the
                            # HistoryDatasetAssociation object.
                            if check.metadata is not None:
                                metadata_value = check.metadata.get(check_attribute, None)
                                if metadata_value is not None:
                                    if str(metadata_value) == str(check_value):
                                        ext = check_format
                                        new_format_set = True
                                        break
            if new_format_set:
                break
    return ext



#!/usr/bin/env python
"""
Provides wrappers and utilities for working with MAF files and alignments.
"""
# Dan Blankenberg

import functools
import logging
import os
import resource
import sys
import tempfile
from copy import deepcopy
from errno import EMFILE

import bx.align.maf
import bx.interval_index_file
import bx.intervals

maketrans = str.maketrans

log = logging.getLogger(__name__)

GAP_CHARS = ["-"]
SRC_SPLIT_CHAR = "."


def src_split(src):
    fields = src.split(SRC_SPLIT_CHAR, 1)
    spec = fields.pop(0)
    if fields:
        chrom = fields.pop(0)
    else:
        chrom = spec
    return spec, chrom


def src_merge(spec, chrom, contig=None):
    if None in [spec, chrom]:
        spec = chrom = spec or chrom
    return bx.align.src_merge(spec, chrom, contig)


def get_species_in_block(block):
    species = []
    for c in block.components:
        spec, chrom = src_split(c.src)
        if spec not in species:
            species.append(spec)
    return species


def tool_fail(msg="Unknown Error"):
    sys.exit(f"Fatal Error: {msg}")


class TempFileHandler:
    """
    Handles creating, opening, closing, and deleting of Temp files, with a
    maximum number of files open at one time.
    """

    DEFAULT_MAX_OPEN_FILES = max(resource.getrlimit(resource.RLIMIT_NOFILE)[0] / 2, 1)

    def __init__(self, max_open_files=None, **kwds):
        if max_open_files is None:
            max_open_files = self.DEFAULT_MAX_OPEN_FILES
        self.max_open_files = max_open_files
        self.files = []
        self.open_file_indexes = []
        self.kwds = kwds

    def get_open_tempfile(self, index=None, **kwds):
        if index is not None and index in self.open_file_indexes:
            self.open_file_indexes.remove(index)
        else:
            if self.max_open_files:
                while len(self.open_file_indexes) >= self.max_open_files:
                    self.close(self.open_file_indexes[0])
            if index is None:
                index = len(self.files)
                temp_kwds = dict(self.kwds)
                temp_kwds["delete"] = False
                temp_kwds.update(kwds)
                # Being able to use delete=True here, would simplify a bit,
                # but we support python2.4 in these tools
                while True:
                    try:
                        tmp_file = tempfile.NamedTemporaryFile(**temp_kwds)
                        filename = tmp_file.name
                        break
                    except OSError as e:
                        if self.open_file_indexes and e.errno == EMFILE:
                            self.max_open_files = len(self.open_file_indexes)
                            self.close(self.open_file_indexes[0])
                        else:
                            raise e
                tmp_file.close()
                self.files.append(open(filename, "r+"))
            else:
                while True:
                    try:
                        self.files[index] = open(self.files[index].name, "r+")
                        break
                    except OSError as e:
                        if self.open_file_indexes and e.errno == EMFILE:
                            self.max_open_files = len(self.open_file_indexes)
                            self.close(self.open_file_indexes[0])
                        else:
                            raise e
                self.files[index].seek(0, 2)
        self.open_file_indexes.append(index)
        return index, self.files[index]

    def close(self, index, delete=False):
        if index in self.open_file_indexes:
            self.open_file_indexes.remove(index)
        rval = self.files[index].close()
        if delete:
            try:
                os.unlink(self.files[index].name)
            except OSError:
                pass
        return rval

    def flush(self, index):
        if index in self.open_file_indexes:
            self.files[index].flush()

    def __del__(self):
        for i in range(len(self.files)):
            self.close(i, delete=True)


# an object corresponding to a reference layered alignment
class RegionAlignment:
    DNA_COMPLEMENT = maketrans("ACGTacgt", "TGCAtgca")
    MAX_SEQUENCE_SIZE = sys.maxsize  # Maximum length of sequence allowed

    def __init__(self, size, species=None, temp_file_handler=None):
        assert (
            size <= self.MAX_SEQUENCE_SIZE
        ), "Maximum length allowed for an individual sequence has been exceeded (%i > %i)." % (
            size,
            self.MAX_SEQUENCE_SIZE,
        )
        species = species or []
        self.size = size
        if not temp_file_handler:
            temp_file_handler = TempFileHandler()
        self.temp_file_handler = temp_file_handler
        self.sequences = {}
        if not isinstance(species, list):
            species = [species]
        for spec in species:
            self.add_species(spec)

    # add a species to the alignment
    def add_species(self, species):
        # make temporary sequence files
        file_index, fh = self.temp_file_handler.get_open_tempfile()
        self.sequences[species] = file_index
        fh.write("-" * self.size)

    # returns the names for species found in alignment, skipping names as requested
    def get_species_names(self, skip=None):
        skip = skip or []
        if not isinstance(skip, list):
            skip = [skip]
        names = list(self.sequences.keys())
        for name in skip:
            try:
                names.remove(name)
            except ValueError:
                pass
        return names

    # returns the sequence for a species
    def get_sequence(self, species):
        file_index, fh = self.temp_file_handler.get_open_tempfile(self.sequences[species])
        fh.seek(0)
        return fh.read()

    # returns the reverse complement of the sequence for a species
    def get_sequence_reverse_complement(self, species):
        complement = [base for base in self.get_sequence(species).translate(self.DNA_COMPLEMENT)]
        complement.reverse()
        return "".join(complement)

    # sets a position for a species
    def set_position(self, index, species, base):
        if len(base) != 1:
            raise Exception("A genomic position can only have a length of 1.")
        return self.set_range(index, species, base)

    # sets a range for a species

    def set_range(self, index, species, bases):
        if index >= self.size or index < 0:
            raise Exception("Your index (%i) is out of range (0 - %i)." % (index, self.size - 1))
        if len(bases) == 0:
            raise Exception("A set of genomic positions can only have a positive length.")
        if species not in self.sequences.keys():
            self.add_species(species)
        file_index, fh = self.temp_file_handler.get_open_tempfile(self.sequences[species])
        fh.seek(index)
        fh.write(bases)

    # Flush temp file of specified species, or all species
    def flush(self, species=None):
        if species is None:
            species = self.sequences.keys()
        elif not isinstance(species, list):
            species = [species]
        for spec in species:
            self.temp_file_handler.flush(self.sequences[spec])


class GenomicRegionAlignment(RegionAlignment):
    def __init__(self, start, end, species=None, temp_file_handler=None):
        species = species or []
        RegionAlignment.__init__(self, end - start, species, temp_file_handler=temp_file_handler)
        self.start = start
        self.end = end


class SplicedAlignment:
    DNA_COMPLEMENT = maketrans("ACGTacgt", "TGCAtgca")

    def __init__(self, exon_starts, exon_ends, species=None, temp_file_handler=None):
        species = species or []
        if not isinstance(exon_starts, list):
            exon_starts = [exon_starts]
        if not isinstance(exon_ends, list):
            exon_ends = [exon_ends]
        assert len(exon_starts) == len(exon_ends), "The number of starts does not match the number of sizes."
        self.exons = []
        if not temp_file_handler:
            temp_file_handler = TempFileHandler()
        self.temp_file_handler = temp_file_handler
        for i in range(len(exon_starts)):
            self.exons.append(
                GenomicRegionAlignment(exon_starts[i], exon_ends[i], species, temp_file_handler=temp_file_handler)
            )

    # returns the names for species found in alignment, skipping names as requested
    def get_species_names(self, skip=None):
        skip = skip or []
        if not isinstance(skip, list):
            skip = [skip]
        names = []
        for exon in self.exons:
            for name in exon.get_species_names(skip=skip):
                if name not in names:
                    names.append(name)
        return names

    # returns the sequence for a species
    def get_sequence(self, species):
        index, fh = self.temp_file_handler.get_open_tempfile()
        for exon in self.exons:
            if species in exon.get_species_names():
                seq = exon.get_sequence(species)
                # we need to refetch fh here, since exon.get_sequence( species ) uses a tempfile
                # and if max==1, it would close fh
                index, fh = self.temp_file_handler.get_open_tempfile(index)
                fh.write(seq)
            else:
                fh.write("-" * exon.size)
        fh.seek(0)
        rval = fh.read()
        self.temp_file_handler.close(index, delete=True)
        return rval

    # returns the reverse complement of the sequence for a species
    def get_sequence_reverse_complement(self, species):
        complement = [base for base in self.get_sequence(species).translate(self.DNA_COMPLEMENT)]
        complement.reverse()
        return "".join(complement)

    # Start and end of coding region
    @property
    def start(self):
        return self.exons[0].start

    @property
    def end(self):
        return self.exons[-1].end


# Open a MAF index using a UID
def maf_index_by_uid(maf_uid, index_location_file):
    for line in open(index_location_file):
        try:
            # read each line, if not enough fields, go to next line
            if line[0:1] == "#":
                continue
            fields = line.split("\t")
            if maf_uid == fields[1]:
                try:
                    maf_files = fields[4].replace("\n", "").replace("\r", "").split(",")
                    return bx.align.maf.MultiIndexed(maf_files, keep_open=True, parse_e_rows=False)
                except Exception as e:
                    raise Exception(f"MAF UID ({maf_uid}) found, but configuration appears to be malformed: {e}")
        except Exception:
            pass
    return None


# return ( index, temp_index_filename ) for user maf, if available, or build one and return it, return None when no tempfile is created
def open_or_build_maf_index(maf_file, index_filename, species=None):
    try:
        return (bx.align.maf.Indexed(maf_file, index_filename=index_filename, keep_open=True, parse_e_rows=False), None)
    except Exception:
        return build_maf_index(maf_file, species=species)


def build_maf_index_species_chromosomes(filename, index_species=None):
    species = []
    species_chromosomes = {}
    indexes = bx.interval_index_file.Indexes()
    blocks = 0
    try:
        maf_reader = bx.align.maf.Reader(open(filename))
        while True:
            pos = maf_reader.file.tell()
            block = next(maf_reader)
            if block is None:
                break
            blocks += 1
            for c in block.components:
                spec = c.src
                chrom = None
                if "." in spec:
                    spec, chrom = spec.split(".", 1)
                if spec not in species:
                    species.append(spec)
                    species_chromosomes[spec] = []
                if chrom and chrom not in species_chromosomes[spec]:
                    species_chromosomes[spec].append(chrom)
                if index_species is None or spec in index_species:
                    forward_strand_start = c.forward_strand_start
                    forward_strand_end = c.forward_strand_end
                    try:
                        forward_strand_start = int(forward_strand_start)
                        forward_strand_end = int(forward_strand_end)
                    except ValueError:
                        continue  # start and end are not integers, can't add component to index, goto next component
                        # this likely only occurs when parse_e_rows is True?
                        # could a species exist as only e rows? should the
                    if forward_strand_end > forward_strand_start:
                        # require positive length; i.e. certain lines have start = end = 0 and cannot be indexed
                        indexes.add(c.src, forward_strand_start, forward_strand_end, pos, max=c.src_size)
    except Exception as e:
        # most likely a bad MAF
        log.debug(f"Building MAF index on {filename} failed: {e}")
        return (None, [], {}, 0)
    return (indexes, species, species_chromosomes, blocks)


# builds and returns ( index, index_filename ) for specified maf_file
def build_maf_index(maf_file, species=None):
    indexes, *_ = build_maf_index_species_chromosomes(maf_file, species)
    if indexes is not None:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as index:
            indexes.write(index)
        return (
            bx.align.maf.Indexed(maf_file, index_filename=index.name, keep_open=True, parse_e_rows=False),
            index.name,
        )
    return (None, None)


def component_overlaps_region(c, region):
    if c is None:
        return False
    start, end = c.get_forward_strand_start(), c.get_forward_strand_end()
    if region.start >= end or region.end <= start:
        return False
    return True


def chop_block_by_region(block, src, region, species=None, mincols=0):
    # This chopping method was designed to maintain consistency with how start/end padding gaps have been working in Galaxy thus far:
    #   behavior as seen when forcing blocks to be '+' relative to src sequence (ref) and using block.slice_by_component( ref, slice_start, slice_end )
    #   whether-or-not this is the 'correct' behavior is questionable, but this will at least maintain consistency
    # comments welcome
    slice_start = block.text_size  # max for the min()
    slice_end = 0  # min for the max()
    old_score = block.score  # save old score for later use
    # We no longer assume only one occurance of src per block, so we need to check them all
    for c in iter_components_by_src(block, src):
        if component_overlaps_region(c, region):
            if c.text is not None:
                rev_strand = False
                if c.strand == "-":
                    # We want our coord_to_col coordinates to be returned from positive stranded component
                    rev_strand = True
                    c = c.reverse_complement()
                start = max(region.start, c.start)
                end = min(region.end, c.end)
                start = c.coord_to_col(start)
                end = c.coord_to_col(end)
                if rev_strand:
                    # need to orient slice coordinates to the original block direction
                    slice_len = end - start
                    end = len(c.text) - start
                    start = end - slice_len
                slice_start = min(start, slice_start)
                slice_end = max(end, slice_end)

    if slice_start < slice_end:
        block = block.slice(slice_start, slice_end)
        if block.text_size > mincols:
            # restore old score, may not be accurate, but it is better than 0 for everything?
            block.score = old_score
            if species is not None:
                block = block.limit_to_species(species)
                block.remove_all_gap_columns()
            return block
    return None


def orient_block_by_region(block, src, region, force_strand=None):
    # loop through components matching src,
    # make sure each of these components overlap region
    # cache strand for each of overlaping regions
    # if force_strand / region.strand not in strand cache, reverse complement
    # we could have 2 sequences with same src, overlapping region, on different strands, this would cause no reverse_complementing
    strands = [c.strand for c in iter_components_by_src(block, src) if component_overlaps_region(c, region)]
    if (
        strands
        and (force_strand is None and region.strand not in strands)
        or (force_strand is not None and force_strand not in strands)
    ):
        block = block.reverse_complement()
    return block


def get_oriented_chopped_blocks_for_region(index, src, region, species=None, mincols=0, force_strand=None):
    for block, _, _ in get_oriented_chopped_blocks_with_index_offset_for_region(
        index, src, region, species, mincols, force_strand
    ):
        yield block


def get_oriented_chopped_blocks_with_index_offset_for_region(
    index, src, region, species=None, mincols=0, force_strand=None
):
    for block, idx, offset in get_chopped_blocks_with_index_offset_for_region(index, src, region, species, mincols):
        yield orient_block_by_region(block, src, region, force_strand), idx, offset


# split a block with multiple occurances of src into one block per src
def iter_blocks_split_by_src(block, src):
    for src_c in iter_components_by_src(block, src):
        new_block = bx.align.Alignment(score=block.score, attributes=deepcopy(block.attributes))
        new_block.text_size = block.text_size
        for c in block.components:
            if c == src_c or c.src != src:
                new_block.add_component(
                    deepcopy(c)
                )  # components have reference to alignment, don't want to lose reference to original alignment block in original components
        yield new_block


# split a block into multiple blocks with all combinations of a species appearing only once per block
def iter_blocks_split_by_species(block, species=None):
    def __split_components_by_species(components_by_species, new_block):
        if components_by_species:
            # more species with components to add to this block
            components_by_species = deepcopy(components_by_species)
            spec_comps = components_by_species.pop(0)
            for c in spec_comps:
                newer_block = deepcopy(new_block)
                newer_block.add_component(deepcopy(c))
                yield from __split_components_by_species(components_by_species, newer_block)
        else:
            # no more components to add, yield this block
            yield new_block

    # divide components by species
    spec_dict = {}
    if not species:
        species = []
        for c in block.components:
            spec, chrom = src_split(c.src)
            if spec not in spec_dict:
                spec_dict[spec] = []
                species.append(spec)
            spec_dict[spec].append(c)
    else:
        for spec in species:
            spec_dict[spec] = []
            for c in iter_components_by_src_start(block, spec):
                spec_dict[spec].append(c)

    empty_block = bx.align.Alignment(
        score=block.score, attributes=deepcopy(block.attributes)
    )  # should we copy attributes?
    empty_block.text_size = block.text_size
    # call recursive function to split into each combo of spec/blocks
    for value in __split_components_by_species(list(spec_dict.values()), empty_block):
        sort_block_components_by_block(value, block)  # restore original component order
        yield value


# generator yielding only chopped and valid blocks for a specified region
def get_chopped_blocks_for_region(index, src, region, species=None, mincols=0):
    for block, _, _ in get_chopped_blocks_with_index_offset_for_region(index, src, region, species, mincols):
        yield block


def get_chopped_blocks_with_index_offset_for_region(index, src, region, species=None, mincols=0):
    for block, idx, offset in index.get_as_iterator_with_index_and_offset(src, region.start, region.end):
        block = chop_block_by_region(block, src, region, species, mincols)
        if block is not None:
            yield block, idx, offset


# returns a filled region alignment for specified regions
def get_region_alignment(
    index,
    primary_species,
    chrom,
    start,
    end,
    strand="+",
    species=None,
    mincols=0,
    overwrite_with_gaps=True,
    temp_file_handler=None,
):
    if species is not None:
        alignment = RegionAlignment(end - start, species, temp_file_handler=temp_file_handler)
    else:
        alignment = RegionAlignment(end - start, primary_species, temp_file_handler=temp_file_handler)
    return fill_region_alignment(
        alignment, index, primary_species, chrom, start, end, strand, species, mincols, overwrite_with_gaps
    )


# reduces a block to only positions exisiting in the src provided
def reduce_block_by_primary_genome(block, species, chromosome, region_start):
    # returns ( startIndex, {species:texts}
    # where texts' contents are reduced to only positions existing in the primary genome
    src = f"{species}.{chromosome}"
    ref = block.get_component_by_src(src)
    start_offset = ref.start - region_start
    species_texts = {}
    for c in block.components:
        species_texts[c.src.split(".")[0]] = list(c.text)
    # remove locations which are gaps in the primary species, starting from the downstream end
    for i in range(len(species_texts[species]) - 1, -1, -1):
        if species_texts[species][i] == "-":
            for text in species_texts.values():
                text.pop(i)
    for spec, text in species_texts.items():
        species_texts[spec] = "".join(text)
    return (start_offset, species_texts)


# fills a region alignment
def fill_region_alignment(
    alignment, index, primary_species, chrom, start, end, strand="+", species=None, mincols=0, overwrite_with_gaps=True
):
    region = bx.intervals.Interval(start, end)
    region.chrom = chrom
    region.strand = strand
    primary_src = f"{primary_species}.{chrom}"

    # Order blocks overlaping this position by score, lowest first
    blocks = []
    for block, idx, offset in index.get_as_iterator_with_index_and_offset(primary_src, start, end):
        score = float(block.score)
        for i in range(0, len(blocks)):
            if score < blocks[i][0]:
                blocks.insert(i, (score, idx, offset))
                break
        else:
            blocks.append((score, idx, offset))

    # gap_chars_tuple = tuple( GAP_CHARS )
    gap_chars_str = "".join(GAP_CHARS)
    # Loop through ordered blocks and layer by increasing score
    for block_dict in blocks:
        for block in iter_blocks_split_by_species(
            block_dict[1].get_at_offset(block_dict[2])
        ):  # need to handle each occurance of sequence in block seperately
            if component_overlaps_region(block.get_component_by_src(primary_src), region):
                block = chop_block_by_region(block, primary_src, region, species, mincols)  # chop block
                block = orient_block_by_region(block, primary_src, region)  # orient block
                start_offset, species_texts = reduce_block_by_primary_genome(block, primary_species, chrom, start)
                for spec, text in species_texts.items():
                    # we should trim gaps from both sides, since these are not positions in this species genome (sequence)
                    text = text.rstrip(gap_chars_str)
                    gap_offset = 0
                    # while text.startswith( gap_chars_tuple ):
                    while True in [
                        text.startswith(gap_char) for gap_char in GAP_CHARS
                    ]:  # python2.4 doesn't accept a tuple for .startswith()
                        gap_offset += 1
                        text = text[1:]
                        if not text:
                            break
                    if text:
                        if overwrite_with_gaps:
                            alignment.set_range(start_offset + gap_offset, spec, text)
                        else:
                            for i, char in enumerate(text):
                                if char not in GAP_CHARS:
                                    alignment.set_position(start_offset + gap_offset + i, spec, char)
    return alignment


# returns a filled spliced region alignment for specified region with start and end lists
def get_spliced_region_alignment(
    index,
    primary_species,
    chrom,
    starts,
    ends,
    strand="+",
    species=None,
    mincols=0,
    overwrite_with_gaps=True,
    temp_file_handler=None,
):
    # create spliced alignment object
    if species is not None:
        alignment = SplicedAlignment(starts, ends, species, temp_file_handler=temp_file_handler)
    else:
        alignment = SplicedAlignment(starts, ends, [primary_species], temp_file_handler=temp_file_handler)
    for exon in alignment.exons:
        fill_region_alignment(
            exon, index, primary_species, chrom, exon.start, exon.end, strand, species, mincols, overwrite_with_gaps
        )
    return alignment


# loop through string array, only return non-commented lines
def line_enumerator(lines, comment_start="#"):
    i = 0
    for line in lines:
        if not line.startswith(comment_start):
            i += 1
            yield (i, line)


# read a GeneBed file, return list of starts, ends, raw fields
def get_starts_ends_fields_from_gene_bed(line):
    # Starts and ends for exons
    starts = []
    ends = []

    fields = line.split()
    # Requires atleast 12 BED columns
    if len(fields) < 12:
        raise Exception(f"Not a proper 12 column BED line ({line}).")
    tx_start = int(fields[1])
    strand = fields[5]
    if strand != "-":
        strand = "+"  # Default strand is +
    cds_start = int(fields[6])
    cds_end = int(fields[7])

    # Calculate and store starts and ends of coding exons
    region_start, region_end = cds_start, cds_end
    exon_starts = list(map(int, fields[11].rstrip(",\n").split(",")))
    exon_starts = [x + tx_start for x in exon_starts]
    exon_ends = list(map(int, fields[10].rstrip(",").split(",")))
    exon_ends = [x + y for x, y in zip(exon_starts, exon_ends)]
    for start, end in zip(exon_starts, exon_ends):
        start = max(start, region_start)
        end = min(end, region_end)
        if start < end:
            starts.append(start)
            ends.append(end)
    return (starts, ends, fields)


def iter_components_by_src(block, src):
    for c in block.components:
        if c.src == src:
            yield c


def get_components_by_src(block, src):
    return [value for value in iter_components_by_src(block, src)]


def iter_components_by_src_start(block, src):
    for c in block.components:
        if c.src.startswith(src):
            yield c


def get_components_by_src_start(block, src):
    return [value for value in iter_components_by_src_start(block, src)]


def sort_block_components_by_block(block1, block2):
    # orders the components in block1 by the index of the component in block2
    # block1 must be a subset of block2
    # occurs in-place
    return block1.components.sort(
        key=functools.cmp_to_key(lambda x, y: block2.components.index(x) - block2.components.index(y))
    )


def get_species_in_maf(maf_filename):
    species = []
    for block in bx.align.maf.Reader(open(maf_filename)):
        for spec in get_species_in_block(block):
            if spec not in species:
                species.append(spec)
    return species


def parse_species_option(species):
    if species:
        species = species.split(",")
        if "None" not in species:
            return species
    return None  # provided species was '', None, or had 'None' in it


def remove_temp_index_file(index_filename):
    try:
        os.unlink(index_filename)
    except Exception:
        pass


# Below are methods to deal with FASTA files


def get_fasta_header(component, attributes=None, suffix=None):
    attributes = attributes or {}
    header = ">%s(%s):%i-%i|" % (
        component.src,
        component.strand,
        component.get_forward_strand_start(),
        component.get_forward_strand_end(),
    )
    for key, value in attributes.items():
        header = f"{header}{key}={value}|"
    if suffix:
        header = f"{header}{suffix}"
    else:
        header = f"{header}{src_split(component.src)[0]}"
    return header


def get_attributes_from_fasta_header(header):
    if not header:
        return {}
    attributes = {}
    header = header.lstrip(">")
    header = header.strip()
    fields = header.split("|")
    try:
        region = fields[0]
        region = region.split("(", 1)
        temp = region[0].split(".", 1)
        attributes["species"] = temp[0]
        if len(temp) == 2:
            attributes["chrom"] = temp[1]
        else:
            attributes["chrom"] = temp[0]
        region = region[1].split(")", 1)
        attributes["strand"] = region[0]
        region = region[1].lstrip(":").split("-")
        attributes["start"] = int(region[0])
        attributes["end"] = int(region[1])
    except Exception:
        # fields 0 is not a region coordinate
        pass
    if len(fields) > 2:
        for i in range(1, len(fields) - 1):
            prop = fields[i].split("=", 1)
            if len(prop) == 2:
                attributes[prop[0]] = prop[1]
    if len(fields) > 1:
        attributes["__suffix__"] = fields[-1]
    return attributes


def iter_fasta_alignment(filename):
    class fastaComponent:
        def __init__(self, species, text=""):
            self.species = species
            self.text = text

        def extend(self, text):
            self.text = self.text + text.replace("\n", "").replace("\r", "").strip()

    # yields a list of fastaComponents for a FASTA file
    f = open(filename, "rb")
    components = []
    # cur_component = None
    while True:
        line = f.readline()
        if not line:
            if components:
                yield components
            return
        line = line.strip()
        if not line:
            if components:
                yield components
            components = []
        elif line.startswith(">"):
            attributes = get_attributes_from_fasta_header(line)
            components.append(fastaComponent(attributes["species"]))
        elif components:
            components[-1].extend(line)



"""
FormDefinition and field factories
"""
# TODO: A FormDefinitionField is closely linked to a form_builder result.
# Can this functionality be further abstracted and merged with form_builder?
from galaxy.model import (
    FormDefinition,
    FormDefinitionCurrent,
)
from galaxy.util import string_as_bool

FORM_TYPES = {f_type.lower(): f_descript for f_type, f_descript in FormDefinition.types.__members__.items()}


class FormDefinitionFactory:
    def __init__(self, form_types, field_type_factories):
        self.form_types = form_types
        self.field_type_factories = field_type_factories

    def new(self, form_type, name, description=None, fields=None, layout=None, form_definition_current=None):
        """
        Return new FormDefinition.
        """
        assert (
            form_type in self.form_types
        ), f"Invalid FormDefinition type ( {form_type} not in {self.form_types.keys()} )"
        assert name, "FormDefinition requires a name"
        if description is None:
            description = ""
        if layout is None:
            layout = []
        if fields is None:
            fields = []
        # Create new FormDefinitionCurrent
        if form_definition_current is None:
            form_definition_current = FormDefinitionCurrent()
        rval = FormDefinition(
            name=name,
            desc=description,
            form_type=self.form_types[form_type],
            form_definition_current=form_definition_current,
            layout=layout,
            fields=fields,
        )
        form_definition_current.latest_form = rval
        return rval

    def from_elem(self, elem, form_definition_current=None):
        """
        Return FormDefinition created from an xml element.
        """
        name = elem.get("name", None)
        description = elem.get("description", None)
        form_type = elem.get("type", None)
        # load layout
        layout = []
        layouts_elem = elem.find("layout")
        if layouts_elem:
            for layout_elem in layouts_elem.findall("grid"):
                layout_name = layout_elem.get("name", None)
                assert layout_name and layout_name not in layout, "Layout grid element requires a unique name."
                layout.append(layout_name)
        # load fields
        fields = []
        fields_elem = elem.find("fields")
        if fields_elem is not None:
            for field_elem in fields_elem.findall("field"):
                field_type = field_elem.get("type")
                assert field_type in self.field_type_factories, f"Invalid form field type ( {field_type} )."
                fields.append(self.field_type_factories[field_type].from_elem(field_elem, layout))
        # create and return new form
        return self.new(
            form_type,
            name,
            description=description,
            fields=fields,
            layout=layout,
            form_definition_current=form_definition_current,
        )


class FormDefinitionFieldFactory:
    type: str

    def __get_stored_field_type(self, **kwds):
        raise Exception("not implemented")

    def new(self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None):
        """
        Return new FormDefinition field.
        """
        rval = {}
        assert name, "Must provide a name"
        rval["name"] = name
        if not label:
            rval["label"] = name
        else:
            rval["label"] = label
        if required:
            rval["required"] = "required"
        else:
            rval["required"] = "optional"
        if helptext is None:
            helptext = ""
        rval["helptext"] = helptext
        if default is None:
            default = ""
        rval["default"] = default
        rval["visible"] = visible
        # if layout is None: #is this needed?
        #    layout = ''
        rval["layout"] = layout
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition created from an xml element.
        """
        name = elem.get("name")
        label = elem.get("label")
        required = string_as_bool(elem.get("required", "false"))
        default = elem.get("value")
        helptext = elem.get("helptext")
        visible = string_as_bool(elem.get("visible", "true"))
        field_layout = elem.get("layout", None)
        if field_layout:
            assert layout and field_layout in layout, f"Invalid layout specified: {field_layout} not in {layout}"
            field_layout = str(
                layout.index(field_layout)
            )  # existing behavior: integer indexes are stored as strings. why?
        return self.new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=field_layout,
        )


class FormDefinitionTextFieldFactory(FormDefinitionFieldFactory):
    type = "text"

    def __get_stored_field_type(self, area):
        if area:
            return "TextArea"
        else:
            return "TextField"

    def new(
        self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None, area=False
    ):
        """
        Return new FormDefinition field.
        """
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type(area)
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type(string_as_bool(elem.get("area", "false")))
        return rval


class FormDefinitionPasswordFieldFactory(FormDefinitionFieldFactory):
    type = "password"

    def __get_stored_field_type(self):
        return "PasswordField"

    def new(
        self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None, area=False
    ):
        """
        Return new FormDefinition field.
        """
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type()
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type()
        return rval


class FormDefinitionAddressFieldFactory(FormDefinitionFieldFactory):
    type = "address"

    def __get_stored_field_type(self):
        return "AddressField"

    def new(self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None):
        """
        Return new FormDefinition field.
        """
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type()
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type()
        return rval


class FormDefinitionWorkflowFieldFactory(FormDefinitionFieldFactory):
    type = "workflow"

    def __get_stored_field_type(self):
        return "WorkflowField"

    def new(self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None):
        """
        Return new FormDefinition field.
        """
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type()
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type()
        return rval


class FormDefinitionWorkflowMappingFieldFactory(FormDefinitionFieldFactory):
    type = "workflowmapping"

    def __get_stored_field_type(self):
        return "WorkflowMappingField"

    def new(self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None):
        """
        Return new FormDefinition field.
        """
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type()
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type()
        return rval


class FormDefinitionHistoryFieldFactory(FormDefinitionFieldFactory):
    type = "history"

    def __get_stored_field_type(self):
        return "HistoryField"

    def new(self, name=None, label=None, required=False, helptext=None, default=None, visible=True, layout=None):
        """
        Return new FormDefinition field.
        """
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type()
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type()
        return rval


class FormDefinitionSelectFieldFactory(FormDefinitionFieldFactory):
    type = "select"

    def __get_stored_field_type(self, checkboxes):
        if checkboxes:
            return "CheckboxField"
        else:
            return "SelectField"

    def new(
        self,
        name=None,
        label=None,
        required=False,
        helptext=None,
        default=None,
        visible=True,
        layout=None,
        options=None,
        checkboxes=False,
    ):
        """
        Return new FormDefinition field.
        """
        options = options or []
        rval = super().new(
            name=name,
            label=label,
            required=required,
            helptext=helptext,
            default=default,
            visible=visible,
            layout=layout,
        )
        rval["type"] = self.__get_stored_field_type(checkboxes)
        if options is None:
            options = []
        rval["selectlist"] = options
        return rval

    def from_elem(self, elem, layout=None):
        """
        Return FormDefinition field created from an xml element.
        """
        rval = super().from_elem(elem, layout=layout)
        rval["type"] = self.__get_stored_field_type(string_as_bool(elem.get("checkboxes", "false")))
        # load select options
        rval["selectlist"] = []
        for select_option in elem.findall("option"):
            value = select_option.get("value", None)
            assert value is not None, 'Must provide a "value" for a select option'
            rval["selectlist"].append(value)
        return rval


field_type_factories = {
    field.type: field()
    for field in (
        FormDefinitionTextFieldFactory,
        FormDefinitionPasswordFieldFactory,
        FormDefinitionAddressFieldFactory,
        FormDefinitionSelectFieldFactory,
        FormDefinitionWorkflowFieldFactory,
        FormDefinitionWorkflowMappingFieldFactory,
        FormDefinitionHistoryFieldFactory,
    )
}

form_factory = FormDefinitionFactory(FORM_TYPES, field_type_factories)



"""
Tool Parameter specific sanitizing.
"""
import logging
import string

import galaxy.util

log = logging.getLogger(__name__)


class ToolParameterSanitizer:
    """
    Handles tool parameter specific sanitizing.

    >>> from galaxy.util import XML
    >>> sanitizer = ToolParameterSanitizer.from_element(XML(
    ... '''
    ... <sanitizer invalid_char="">
    ...   <valid initial="string.ascii_letters"/>
    ... </sanitizer>
    ... '''))
    >>> sanitizer.sanitize_param(''.join(sorted(c for c in string.printable))) == ''.join(sorted(c for c in string.ascii_letters))
    True
    >>> slash = chr(92)
    >>> sanitizer = ToolParameterSanitizer.from_element(XML(
    ... '''
    ... <sanitizer>
    ...   <valid initial="none">
    ...    <add preset="string.printable"/>
    ...    <remove value="&quot;"/>
    ...    <remove value="%s"/>
    ...   </valid>
    ...   <mapping initial="none">
    ...     <add source="&quot;" target="%s&quot;"/>
    ...     <add source="%s" target="%s%s"/>
    ...   </mapping>
    ... </sanitizer>
    ... ''' % (slash, slash, slash, slash, slash)))
    >>> text = '%s"$rm&#!' % slash
    >>> [c for c in sanitizer.sanitize_param(text)] == [slash, slash, slash, '"', '$', 'r', 'm', '&', '#', '!']
    True
    """

    VALID_PRESET = {"default": (f"{string.ascii_letters + string.digits} -=_.()/+*^,:?!"), "none": ""}
    MAPPING_PRESET = {"default": galaxy.util.mapped_chars, "none": {}}
    DEFAULT_INVALID_CHAR = "X"

    # class methods
    @classmethod
    def from_element(cls, elem):
        """Loads the proper filter by the type attribute of elem"""
        # TODO: Add ability to generically specify a method to use for sanitizing input via specification in tool XML
        rval = ToolParameterSanitizer()
        rval._invalid_char = elem.get("invalid_char", cls.DEFAULT_INVALID_CHAR)
        rval.sanitize = galaxy.util.string_as_bool(elem.get("sanitize", "True"))
        for valid_elem in elem.findall("valid"):
            rval._valid_chars = rval.get_valid_by_name(valid_elem.get("initial", "default"))
            for action_elem in valid_elem:
                preset = rval.get_valid_by_name(action_elem.get("preset", "none"))
                valid_value = [val for val in action_elem.get("value", [])]
                if action_elem.tag.lower() == "add":
                    for val in preset + valid_value:
                        if val not in rval._valid_chars:
                            rval._valid_chars.append(val)
                elif action_elem.tag.lower() == "remove":
                    for val in preset + valid_value:
                        while val in rval._valid_chars:
                            rval._valid_chars.remove(val)
                else:
                    log.debug(f"Invalid action tag in valid: {action_elem.tag}")
        for mapping_elem in elem.findall("mapping"):
            rval._mapped_chars = rval.get_mapping_by_name(mapping_elem.get("initial", "default"))
            for action_elem in mapping_elem:
                map_source = action_elem.get("source", None)
                map_target = action_elem.get("target", None)
                preset = rval.get_mapping_by_name(action_elem.get("preset", "none"))
                if action_elem.tag.lower() == "add":
                    rval._mapped_chars.update(preset)
                    if None not in [map_source, map_target]:
                        rval._mapped_chars[map_source] = map_target
                elif action_elem.tag.lower() == "remove":
                    for map_key in preset.keys():
                        if map_key in rval._mapped_chars:
                            del rval._mapped_chars[map_key]
                    if map_source is not None and map_key in rval._mapped_chars:
                        del rval._mapped_chars[map_key]
                else:
                    log.debug(f"Invalid action tag in mapping: {action_elem.tag}")
        return rval

    @classmethod
    def get_valid_by_name(cls, name):
        rval = []
        for split_name in name.split(","):
            # Remove ';' (if present) and everything after it
            split_name = split_name.split(";", 1)[0]
            split_name = split_name.strip()
            value = []
            if split_name.startswith("string."):
                string_constant = split_name[7:]
                if string_constant in ("letters", "lowercase", "uppercase"):
                    split_name = f"string.ascii_{string_constant}"
                try:
                    value = eval(split_name)
                except NameError as e:
                    log.debug(f"Invalid string preset specified: {e}")
            elif split_name in cls.VALID_PRESET:
                value = cls.VALID_PRESET[split_name]
            else:
                log.debug(f"Invalid preset name specified: {split_name}")
            rval.extend([val for val in value if val not in rval])
        return rval

    @classmethod
    def get_mapping_by_name(cls, name):
        rval = {}
        for split_name in name.split(","):
            split_name = split_name.strip()
            if split_name in cls.MAPPING_PRESET:
                rval.update(cls.MAPPING_PRESET[split_name])
            else:
                log.debug(f"Invalid preset name specified: {split_name}")
        return rval

    # end class methods

    def __init__(self):
        self._valid_chars = []  # List of valid characters
        self._mapped_chars = {}  # Replace a char with a any number of characters
        self._invalid_char = self.DEFAULT_INVALID_CHAR  # Replace invalid characters with this character
        self.sanitize = True  # Simply pass back the passed in value

    def restore_text(self, text):
        """Restores sanitized text"""
        if self.sanitize:
            for key, value in self._mapped_chars.items():
                text = text.replace(value, key)
        return text

    def sanitize_text(self, text):
        """Restricts the characters that are allowed in a text"""
        if not self.sanitize:
            return text
        rval = []
        for c in text:
            if c in self._valid_chars:
                rval.append(c)
            elif c in self._mapped_chars:
                rval.append(self._mapped_chars[c])
            else:
                rval.append(self._invalid_char)
        return "".join(rval)

    def sanitize_param(self, value):
        """Clean incoming parameters (strings or lists)"""
        if not self.sanitize:
            return value
        if isinstance(value, str):
            return self.sanitize_text(value)
        elif isinstance(value, list):
            return list(map(self.sanitize_text, value))
        else:
            raise Exception(f"Unknown parameter type ({type(value)}:{value})")
